# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `distributions.py`."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import distributions


@chex.dataclass(frozen=True)
class _MockActionSpec:
  minimum: chex.Array
  maximum: chex.Array


class CategoricalSampleTest(parameterized.TestCase):

  @chex.all_variants()
  def test_categorical_sample(self):
    key = np.array([1, 2], dtype=np.uint32)
    probs = np.array([0.2, 0.3, 0.5])
    sample = self.variant(distributions.categorical_sample)(key, probs)
    self.assertEqual(sample, 0)

  @chex.all_variants()
  @parameterized.parameters(
      ((-1., 10., -1.),),
      ((0., 0., 0.),),
      ((1., np.inf, 3.),),
      ((1., 2., -np.inf),),
      ((1., 2., np.nan),),
  )
  def test_categorical_sample_on_invalid_distributions(self, probs):
    key = np.array([1, 2], dtype=np.uint32)
    probs = np.asarray(probs)
    sample = self.variant(distributions.categorical_sample)(key, probs)
    self.assertEqual(sample, -1)


class SoftmaxTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.logits = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    self.samples = np.array([0, 1], dtype=np.int32)

    self.expected_probs = np.array(  # softmax with temperature=10
        [[0.34425336, 0.34425336, 0.31149334],
         [0.332225, 0.3671654, 0.3006096]],
        dtype=np.float32)
    probs = np.array(  # softmax with temperature=1
        [[0.42231882, 0.42231882, 0.15536241],
         [0.24472848, 0.66524094, 0.09003057]],
        dtype=np.float32)
    logprobs = np.log(probs)
    self.expected_logprobs = np.array(
        [logprobs[0][self.samples[0]], logprobs[1][self.samples[1]]])
    self.expected_entropy = -np.sum(probs * logprobs, axis=-1)
    self.expected_clipped_entropy = {0.5: 0.549306, 0.9: 0.988751}

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_probs(self, softmax_dist):
    """Tests for a single element."""
    distrib = softmax_dist(temperature=10.)
    softmax = self.variant(distrib.probs)
    # For each element in the batch.
    for logits, expected in zip(self.logits, self.expected_probs):
      # Test outputs.
      actual = softmax(logits)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_probs_batch(self, softmax_dist):
    """Tests for a full batch."""
    distrib = softmax_dist(temperature=10.)
    softmax = self.variant(distrib.probs)
    # Test softmax output in batch.
    actual = softmax(self.logits)
    np.testing.assert_allclose(self.expected_probs, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_logprob(self, softmax_dist):
    """Tests for a single element."""
    distrib = softmax_dist()
    logprob_fn = self.variant(distrib.logprob)
    # For each element in the batch.
    for logits, samples, expected in zip(
        self.logits, self.samples, self.expected_logprobs):
      # Test output.
      actual = logprob_fn(samples, logits)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_logprob_batch(self, softmax_dist):
    """Tests for a full batch."""
    distrib = softmax_dist()
    logprob_fn = self.variant(distrib.logprob)
    # Test softmax output in batch.
    actual = logprob_fn(self.samples, self.logits)
    np.testing.assert_allclose(self.expected_logprobs, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_entropy(self, softmax_dist):
    """Tests for a single element."""
    distrib = softmax_dist()
    entropy_fn = self.variant(distrib.entropy)
    # For each element in the batch.
    for logits, expected in zip(self.logits, self.expected_entropy):
      # Test outputs.
      actual = entropy_fn(logits)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.parameters((0.9, [0.988751, 0.832396]),
                            (0.5, [0.549306, 0.549306]))
  def test_softmax_clipped_entropy_batch(self, entropy_clip, expected_clipped):
    """Tests for a single element."""
    distrib = distributions.clipped_entropy_softmax(entropy_clip=entropy_clip)
    entropy_fn = self.variant(distrib.entropy)
    # Test softmax output in batch.
    actual = entropy_fn(self.logits)
    np.testing.assert_allclose(expected_clipped, actual, atol=1e-4)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('softmax', distributions.softmax),
      ('clipped_entropy_softmax', distributions.clipped_entropy_softmax))
  def test_softmax_entropy_batch(self, softmax_dist):
    """Tests for a full batch."""
    distrib = softmax_dist()
    entropy_fn = self.variant(distrib.entropy)
    # Test softmax output in batch.
    actual = entropy_fn(self.logits)
    np.testing.assert_allclose(self.expected_entropy, actual, atol=1e-4)


class GreedyTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.preferences = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    self.samples = np.array([0, 1], dtype=np.int32)

    self.expected_probs = np.array(
        [[0.5, 0.5, 0.], [0., 1., 0.]], dtype=np.float32)
    self.expected_logprob = np.array(
        [-0.6931472, 0.], dtype=np.float32)
    self.expected_entropy = np.array(
        [0.6931472, 0.], dtype=np.float32)

  @chex.all_variants()
  def test_greedy_probs(self):
    """Tests for a single element."""
    distrib = distributions.greedy()
    greedy = self.variant(distrib.probs)
    # For each element in the batch.
    for preferences, expected in zip(self.preferences, self.expected_probs):
      # Test outputs.
      actual = greedy(preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_probs_batch(self):
    """Tests for a full batch."""
    distrib = distributions.greedy()
    greedy = self.variant(distrib.probs)
    # Test greedy output in batch.
    actual = greedy(self.preferences)
    np.testing.assert_allclose(self.expected_probs, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_logprob(self):
    """Tests for a single element."""
    distrib = distributions.greedy()
    logprob_fn = self.variant(distrib.logprob)
    # For each element in the batch.
    for preferences, samples, expected in zip(
        self.preferences, self.samples, self.expected_logprob):
      # Test output.
      actual = logprob_fn(samples, preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_logprob_batch(self):
    """Tests for a full batch."""
    distrib = distributions.greedy()
    logprob_fn = self.variant(distrib.logprob)
    # Test greedy output in batch.
    actual = logprob_fn(self.samples, self.preferences)
    np.testing.assert_allclose(self.expected_logprob, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_entropy(self):
    """Tests for a single element."""
    distrib = distributions.greedy()
    entropy_fn = self.variant(distrib.entropy)
    # For each element in the batch.
    for preferences, expected in zip(self.preferences, self.expected_entropy):
      # Test outputs.
      actual = entropy_fn(preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_entropy_batch(self):
    """Tests for a full batch."""
    distrib = distributions.greedy()
    entropy_fn = self.variant(distrib.entropy)
    # Test greedy output in batch.
    actual = entropy_fn(self.preferences)
    np.testing.assert_allclose(self.expected_entropy, actual, atol=1e-4)


class EpsilonGreedyTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.epsilon = 0.2

    self.preferences = np.array([[1, 1, 0, 0], [1, 2, 0, 0]], dtype=np.float32)
    self.samples = np.array([0, 1], dtype=np.int32)

    self.expected_probs = np.array(
        [[0.45, 0.45, 0.05, 0.05], [0.05, 0.85, 0.05, 0.05]], dtype=np.float32)
    self.expected_logprob = np.array(
        [-0.7985077, -0.1625189], dtype=np.float32)
    self.expected_entropy = np.array(
        [1.01823008, 0.58750093], dtype=np.float32)

  @chex.all_variants()
  def test_greedy_probs(self):
    """Tests for a single element."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    probs_fn = self.variant(distrib.probs)
    # For each element in the batch.
    for preferences, expected in zip(self.preferences, self.expected_probs):
      # Test outputs.
      actual = probs_fn(preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_probs_batch(self):
    """Tests for a full batch."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    probs_fn = self.variant(distrib.probs)
    # Test greedy output in batch.
    actual = probs_fn(self.preferences)
    np.testing.assert_allclose(self.expected_probs, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_logprob(self):
    """Tests for a single element."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    logprob_fn = self.variant(distrib.logprob)
    # For each element in the batch.
    for preferences, samples, expected in zip(
        self.preferences, self.samples, self.expected_logprob):
      # Test output.
      actual = logprob_fn(samples, preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_logprob_batch(self):
    """Tests for a full batch."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    logprob_fn = self.variant(distrib.logprob)
    # Test greedy output in batch.
    actual = logprob_fn(self.samples, self.preferences)
    np.testing.assert_allclose(self.expected_logprob, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_entropy(self):
    """Tests for a single element."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    entropy_fn = self.variant(distrib.entropy)
    # For each element in the batch.
    for preferences, expected in zip(self.preferences, self.expected_entropy):
      # Test outputs.
      actual = entropy_fn(preferences)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_greedy_entropy_batch(self):
    """Tests for a full batch."""
    distrib = distributions.epsilon_greedy(self.epsilon)
    entropy_fn = self.variant(distrib.entropy)
    # Test greedy output in batch.
    actual = entropy_fn(self.preferences)
    np.testing.assert_allclose(self.expected_entropy, actual, atol=1e-4)


class GaussianDiagonalTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.mu = np.array([[1., -1], [0.1, -0.1]], dtype=np.float32)
    self.sigma = np.array([[0.1, 0.1], [0.2, 0.3]], dtype=np.float32)
    self.sample = np.array([[1.2, -1.1], [-0.1, 0.]], dtype=np.float32)
    self.other_mu = np.array([[1., -10.], [0.3, -0.2]], dtype=np.float32)
    self.other_sigma = np.array([[0.1, 0.1], [0.8, 0.3]], dtype=np.float32)

    # Expected values for the distribution's function were computed using
    # tfd.MultivariateNormalDiag (from the tensorflow_probability package).
    self.expected_prob_a = np.array(
        [1.3064219, 1.5219283], dtype=np.float32)
    self.expected_logprob_a = np.array(
        [0.26729202, 0.41997814], dtype=np.float32)
    self.expected_entropy = np.array(
        [-1.7672932, 0.02446628], dtype=np.float32)
    self.expected_kl = np.array(
        [4050.00, 1.00435], dtype=np.float32)
    self.expected_kl_to_std_normal = np.array(
        [4.6151705, 1.8884108], dtype=np.float32)

  @chex.all_variants()
  def test_gaussian_prob(self):
    """Tests for a single element."""
    distrib = distributions.gaussian_diagonal()
    prob_fn = self.variant(distrib.prob)
    # For each element in the batch.
    for mu, sigma, sample, expected in zip(
        self.mu, self.sigma, self.sample, self.expected_prob_a):
      # Test outputs.
      actual = prob_fn(sample, mu, sigma)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_prob_batch(self):
    """Tests for a full batch."""
    distrib = distributions.gaussian_diagonal()
    prob_fn = self.variant(distrib.prob)
    # Test greedy output in batch.
    actual = prob_fn(self.sample, self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_prob_a, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_logprob(self):
    """Tests for a single element."""
    distrib = distributions.gaussian_diagonal()
    logprob_fn = self.variant(distrib.logprob)
    # For each element in the batch.
    for mu, sigma, sample, expected in zip(
        self.mu, self.sigma, self.sample, self.expected_logprob_a):
      # Test output.
      actual = logprob_fn(sample, mu, sigma)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_logprob_batch(self):
    """Tests for a full batch."""
    distrib = distributions.gaussian_diagonal()
    logprob_fn = self.variant(distrib.logprob)
    # Test greedy output in batch.
    actual = logprob_fn(self.sample, self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_logprob_a, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_entropy(self):
    """Tests for a single element."""
    distrib = distributions.gaussian_diagonal()
    entropy_fn = self.variant(distrib.entropy)
    # For each element in the batch.
    for mu, sigma, expected in zip(
        self.mu, self.sigma, self.expected_entropy):
      # Test outputs.
      actual = entropy_fn(mu, sigma)
      np.testing.assert_allclose(expected, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_entropy_batch(self):
    """Tests for a full batch."""
    distrib = distributions.gaussian_diagonal()
    entropy_fn = self.variant(distrib.entropy)
    # Test greedy output in batch.
    actual = entropy_fn(self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_entropy, actual, atol=1e-4)

  @chex.all_variants()
  def test_gaussian_kl_batch(self):
    """Tests for a full batch."""
    distrib = distributions.gaussian_diagonal()
    kl_fn = self.variant(distrib.kl)
    # Test greedy output in batch.
    actual = kl_fn(self.mu, self.sigma, self.other_mu, self.other_sigma)
    np.testing.assert_allclose(self.expected_kl, actual, atol=1e-3, rtol=1e-6)

  @chex.all_variants()
  def test_gaussian_kl_to_std_normal_batch(self):
    """Tests for a full batch."""
    distrib = distributions.gaussian_diagonal()
    kl_fn = self.variant(distrib.kl_to_standard_normal)
    # Test greedy output in batch.
    actual = kl_fn(self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_kl_to_std_normal, actual,
                               atol=1e-4)


class SquashedGaussianTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.mu = np.array([[1., -1], [0.1, -0.1]], dtype=np.float32)
    self.sigma = np.array([[0.1, 0.1], [0.2, 0.3]], dtype=np.float32)
    self.sample = np.array([[0.5, -0.6], [-.4, -.2]], dtype=np.float32)
    self.other_mu = np.array([[1., -10.], [0.3, -0.2]], dtype=np.float32)
    self.other_sigma = np.array([[0.1, 0.1], [0.8, 0.3]], dtype=np.float32)
    self.action_spec = _MockActionSpec(minimum=np.array([-1.0]),
                                       maximum=np.array([2.0]))
    self.sigma_min = 0.0
    self.sigma_max = 2.0
    # Expected values for the distribution's function were computed using
    # tfd.MultivariateNormalDiag (from the tensorflow_probability package).
    self.expected_prob_a = np.array(
        [0.016403, 0.011328], dtype=np.float32)
    self.expected_logprob_a = np.array(
        [-4.110274, -4.480485], dtype=np.float32)
    self.expected_entropy = np.array(
        [5.037213, 5.326565], dtype=np.float32)
    self.expected_kl = np.array(
        [0.003151, 0.164303], dtype=np.float32)
    self.expected_kl_to_std_normal = np.array(
        [6.399713, 8.61989], dtype=np.float32)

  @chex.all_variants()
  def test_squashed_gaussian_prob(self):
    """Tests for a full batch."""
    distrib = distributions.squashed_gaussian(sigma_min=self.sigma_min,
                                              sigma_max=self.sigma_max)
    prob_fn = self.variant(distrib.prob)
    # Test greedy output in batch.
    actual = prob_fn(self.sample, self.mu, self.sigma, self.action_spec)
    np.testing.assert_allclose(self.expected_prob_a, actual, atol=1e-4)

  @chex.all_variants()
  def test_squashed_gaussian_logprob(self):
    """Tests for a full batch."""
    distrib = distributions.squashed_gaussian(sigma_min=self.sigma_min,
                                              sigma_max=self.sigma_max)
    logprob_fn = self.variant(distrib.logprob)
    # Test greedy output in batch.
    actual = logprob_fn(self.sample, self.mu, self.sigma, self.action_spec)
    np.testing.assert_allclose(self.expected_logprob_a, actual, atol=1e-3,
                               rtol=1e-6)

  @chex.all_variants()
  def test_squashed_gaussian_entropy(self):
    """Tests for a full batch."""
    distrib = distributions.squashed_gaussian(sigma_min=self.sigma_min,
                                              sigma_max=self.sigma_max)
    entropy_fn = self.variant(distrib.entropy)
    # Test greedy output in batch.
    actual = entropy_fn(self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_entropy, actual, atol=1e-3,
                               rtol=1e-6)

  @chex.all_variants()
  def test_squashed_gaussian_kl(self):
    """Tests for a full batch."""
    distrib = distributions.squashed_gaussian(sigma_min=self.sigma_min,
                                              sigma_max=self.sigma_max)
    kl_fn = self.variant(distrib.kl)
    # Test greedy output in batch.
    actual = kl_fn(self.mu, self.sigma, self.other_mu, self.other_sigma)
    np.testing.assert_allclose(self.expected_kl, actual, atol=1e-3, rtol=1e-6)

  @chex.all_variants()
  def test_squashed_gaussian_kl_to_std_normal(self):
    """Tests for a full batch."""
    distrib = distributions.squashed_gaussian(sigma_min=self.sigma_min,
                                              sigma_max=self.sigma_max)
    kl_fn = self.variant(distrib.kl_to_standard_normal)
    # Test greedy output in batch.
    actual = kl_fn(self.mu, self.sigma)
    np.testing.assert_allclose(self.expected_kl_to_std_normal, actual,
                               atol=1e-3, rtol=1e-5)


class ImportanceSamplingTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.pi_logits = np.array([[0.2, 0.8], [0.6, 0.4]], dtype=np.float32)
    self.mu_logits = np.array([[0.8, 0.2], [0.6, 0.4]], dtype=np.float32)
    self.actions = np.array([1, 0], dtype=np.int32)

    pi = jax.nn.softmax(self.pi_logits)
    mu = jax.nn.softmax(self.mu_logits)
    self.expected_rhos = np.array(
        [pi[0][1] / mu[0][1], pi[1][0] / mu[1][0]], dtype=np.float32)

  @chex.all_variants()
  def test_importance_sampling_ratios_batch(self):
    """Tests for a full batch."""
    ratios_fn = self.variant(
        distributions.categorical_importance_sampling_ratios)
    # Test softmax output in batch.
    actual = ratios_fn(self.pi_logits, self.mu_logits, self.actions)
    np.testing.assert_allclose(self.expected_rhos, actual, atol=1e-4)


class CategoricalKLTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.p_logits = np.array([[1, 1, 0], [1, 2, 0]], dtype=np.float32)
    p_probs = np.array([[0.42231882, 0.42231882, 0.15536241],
                        [0.24472848, 0.66524094, 0.09003057]],
                       dtype=np.float32)
    p_logprobs = np.log(p_probs)
    self.q_logits = np.array([[1, 2, 0], [1, 1, 0]], dtype=np.float32)
    q_probs = np.array([[0.24472848, 0.66524094, 0.09003057],
                        [0.42231882, 0.42231882, 0.15536241]],
                       dtype=np.float32)
    q_logprobs = np.log(q_probs)

    self.expected_kl = np.sum(p_probs * (p_logprobs - q_logprobs), axis=-1)

  @chex.all_variants()
  def test_categorical_kl_divergence_batch(self):
    """Tests for a full batch."""
    kl_fn = self.variant(distributions.categorical_kl_divergence)
    # Test softmax output in batch.
    actual = kl_fn(self.p_logits, self.q_logits)
    np.testing.assert_allclose(self.expected_kl, actual, atol=1e-4)


class CategoricalCrossEntropyTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()

    self.labels = np.array([[0., 1., 0.], [1., 0., 0.]], dtype=np.float32)
    self.logits = np.array([[10., 1., -2.], [1., 4., 0.2]], dtype=np.float32)

    self.expected = np.array([9.00013, 3.0696733], dtype=np.float32)

  @chex.all_variants()
  def test_categorical_cross_entropy_batch(self):
    """Tests for a full batch."""
    cross_entropy = self.variant(jax.vmap(
        distributions.categorical_cross_entropy))
    # Test outputs.
    actual = cross_entropy(self.labels, self.logits)
    np.testing.assert_allclose(self.expected, actual, atol=1e-4)


class MultivariateNormalKLTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    # Test numbers taken from tfd.MultivariateNormalDiag
    self.mu0 = np.array([[5., -1], [0.1, -0.1]], dtype=np.float32)
    self.sigma0 = np.array([[0.3, 0.1], [0.2, 0.3]], dtype=np.float32)
    self.mu1 = np.array([[0.005, -11.], [-0.25, -0.2]], dtype=np.float32)
    self.sigma1 = np.array([[0.1, 0.1], [0.6, 0.3]], dtype=np.float32)
    self.expected_kl = np.array([6.2504023e+03, 8.7986231e-01],
                                dtype=np.float32)

  @chex.all_variants()
  def test_multivariate_normal_kl_divergence_batch(self):
    kl_fn = self.variant(distributions.multivariate_normal_kl_divergence)
    actual = kl_fn(self.mu0, self.sigma0, self.mu1, self.sigma1)
    np.testing.assert_allclose(self.expected_kl, actual, atol=1e-3, rtol=1e-6)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for exploration.py."""

from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import exploration


class GaussianTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._num_actions = 3
    self._rng_key = jax.random.PRNGKey(42)

  @chex.all_variants()
  def test_deterministic(self):
    """Check that noisy and noisless actions match for zero stddev."""
    add_noise = self.variant(exploration.add_gaussian_noise)
    # Test that noisy and noisless actions match for zero stddev
    for _ in range(10):
      action = np.random.normal(0., 1., self._num_actions)
      # Test output.
      self._rng_key, key = jax.random.split(self._rng_key)
      noisy_action = add_noise(key, action, 0.)
      np.testing.assert_allclose(action, noisy_action)


class OrnsteinUhlenbeckTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._num_actions = 3
    self._rng_key = jax.random.PRNGKey(42)

  @chex.all_variants()
  def test_deterministic(self):
    """Check that noisy and noisless actions match for zero stddev."""
    add_noise = self.variant(exploration.add_ornstein_uhlenbeck_noise)
    # Test that noisy and noisless actions match for zero stddev
    noise_tm1 = np.zeros((self._num_actions,))
    for _ in range(10):
      action = np.random.normal(0., 1., self._num_actions)
      # Test output.
      self._rng_key, key = jax.random.split(self._rng_key)
      noisy_action = add_noise(key, action, noise_tm1, 1., 0.)
      noise_tm1 = action - noisy_action
      np.testing.assert_allclose(action, noisy_action)


class DirichletNoiseTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._batch_size = 5
    self._num_actions = 10
    self._rng_key = jax.random.PRNGKey(42)

  @chex.all_variants()
  def test_deterministic(self):
    """Check that noisy and noisless actions match for zero stddev."""
    add_noise = self.variant(exploration.add_dirichlet_noise)

    # Test that noisy and noisless actions match for zero Dirichlet noise
    for _ in range(10):
      prior = np.random.normal(0., 1., (self._batch_size, self._num_actions))

      # Test output.
      self._rng_key, key = jax.random.split(self._rng_key)
      noisy_prior = add_noise(
          key, prior, dirichlet_alpha=0.3, dirichlet_fraction=0.)
      np.testing.assert_allclose(prior, noisy_prior)


class EMIntrinsicRewardTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.num_neighbors = 2
    self.reward_scale = 1.

  @chex.all_variants()
  def test_novelty_reward(self):
    """Check reward is higher for novel embed than those identical to memory."""

    @self.variant
    def episodic_memory_intrinsic_rewards(embeddings, reward_scale):
      return exploration.episodic_memory_intrinsic_rewards(
          embeddings, self.num_neighbors, reward_scale, max_memory_size=10)
    # Memory starts out as all zeros, if we try to add more zeros we should get
    # a lower reward than if we try to add 2 novel embeddings.
    identical_embeddings = np.array([[0., 0.], [0., 0.]])
    novel_embeddings = np.array([[1.3, 2.7], [-10.4, 16.01]])
    low_reward, state = episodic_memory_intrinsic_rewards(
        identical_embeddings, self.reward_scale)
    np.testing.assert_equal(np.array(state.distance_sum), 0)
    high_reward, _ = episodic_memory_intrinsic_rewards(
        novel_embeddings, self.reward_scale)
    np.testing.assert_array_less(low_reward, high_reward)

  @chex.all_variants()
  def test_custom_memory(self):
    """Check that embeddings are added appropriately to a custom memory."""

    @self.variant
    def episodic_memory_intrinsic_rewards(embeddings, memory, reward_scale):
      return exploration.episodic_memory_intrinsic_rewards(
          embeddings, self.num_neighbors, reward_scale,
          exploration.IntrinsicRewardState(memory=memory, next_memory_index=2),
          max_memory_size=4)

    embeddings = np.array([[2., 2.], [3., 3.], [4., 4.]])
    memory = np.array([[-1., -1.,], [1., 1.], [0., 0.], [0., 0.]])
    _, intrinsic_reward_state = episodic_memory_intrinsic_rewards(
        embeddings, memory, self.reward_scale)

    np.testing.assert_array_equal(
        intrinsic_reward_state.memory,
        # Embeddings should have been added in a ring buffer way.
        np.array([[4., 4.,], [1., 1.], [2., 2.], [3., 3.]]))


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Experiment loop."""

import haiku as hk
import jax


def run_loop(
    agent, environment, accumulator, seed,
    batch_size, train_episodes, evaluate_every, eval_episodes):
  """A simple run loop for examples of reinforcement learning with rlax."""

  # Init agent.
  rng = hk.PRNGSequence(jax.random.PRNGKey(seed))
  params = agent.initial_params(next(rng))
  learner_state = agent.initial_learner_state(params)

  print(f"Training agent for {train_episodes} episodes")
  for episode in range(train_episodes):

    # Prepare agent, environment and accumulator for a new episode.
    timestep = environment.reset()
    accumulator.push(timestep, None)
    actor_state = agent.initial_actor_state()

    while not timestep.last():

      # Acting.
      actor_output, actor_state = agent.actor_step(
          params, timestep, actor_state, next(rng), evaluation=False)

      # Agent-environment interaction.
      action = int(actor_output.actions)
      timestep = environment.step(action)

      # Accumulate experience.
      accumulator.push(timestep, action)

      # Learning.
      if accumulator.is_ready(batch_size):
        params, learner_state = agent.learner_step(
            params, accumulator.sample(batch_size), learner_state, next(rng))

    # Evaluation.
    if not episode % evaluate_every:
      returns = 0.
      for _ in range(eval_episodes):
        timestep = environment.reset()
        actor_state = agent.initial_actor_state()

        while not timestep.last():
          actor_output, actor_state = agent.actor_step(
              params, timestep, actor_state, next(rng), evaluation=True)
          timestep = environment.step(int(actor_output.actions))
          returns += timestep.reward

      avg_returns = returns / eval_episodes
      print(f"Episode {episode:4d}: Average returns: {avg_returns:.2f}")



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `policy_targets.py`."""

from absl.testing import absltest
import distrax
import jax
import jax.numpy as jnp
import numpy as np

from rlax._src import policy_targets


class PolicyTargetsTest(absltest.TestCase):

  def test_sampled_policy_distillation_loss(self):
    targets = policy_targets.PolicyTarget(
        actions=jnp.array([0, 1], dtype=jnp.int32),
        weights=jnp.array([0.5, 0.0], dtype=jnp.float32))
    distribution = distrax.Categorical(
        probs=jnp.array([[0.7, 0.3], [0.9, 0.1]], dtype=jnp.float32))
    loss = policy_targets.sampled_policy_distillation_loss(
        distribution=distribution,
        policy_targets=targets)
    expected_loss = 0.089169
    np.testing.assert_allclose(expected_loss, loss, atol=1e-4)

  def test_constant_policy_targets(self):
    rng_key = jax.random.PRNGKey(42)
    num_samples = 4
    weights_scale = 0.2
    distribution = distrax.Categorical(
        probs=jnp.array([0.5, 0.5], dtype=jnp.float32))
    constant_targets = policy_targets.constant_policy_targets(
        distribution, rng_key, num_samples, weights_scale)
    expected_random_actions = distribution.sample(
        seed=rng_key, sample_shape=(num_samples,))
    expected_target_weights = weights_scale * jnp.ones((num_samples,))
    np.testing.assert_allclose(
        constant_targets.weights,
        expected_target_weights, atol=1e-4)
    np.testing.assert_allclose(
        constant_targets.actions,
        expected_random_actions, atol=1e-4)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tree utilities."""

import functools
from typing import Any, Callable, Sequence

import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import base

Array = chex.Array
tree_structure = jax.tree_util.tree_structure


def tree_select(pred: Array, on_true: Any, on_false: Any):
  """Select either one of two identical nested structs based on condition.

  Args:
    pred: a boolean condition.
    on_true: an arbitrary nested structure.
    on_false: a nested structure identical to `on_true`.

  Returns:
    the selected nested structure.
  """
  if tree_structure(on_true) != tree_structure(on_false):
    raise ValueError('The two branches must have the same structure.')
  return jax.tree_util.tree_map(lambda x, y: jax.lax.select(pred, x, y),
                                     on_true, on_false)


def tree_map_zipped(fn: Callable[..., Any], nests: Sequence[Any]):
  """Map a function over a list of identical nested structures.

  Args:
    fn: the function to map; must have arity equal to `len(list_of_nests)`.
    nests: a list of identical nested structures.

  Returns:
    a nested structure whose leaves are outputs of applying `fn`.
  """
  if not nests:
    return nests
  tree_def = tree_structure(nests[0])
  if any(tree_structure(x) != tree_def for x in nests[1:]):
    raise ValueError('All elements must share the same tree structure.')
  return jax.tree_unflatten(
      tree_def, [fn(*d) for d in zip(*[jax.tree_leaves(x) for x in nests])])


def tree_split_key(rng_key: Array, tree_like: Any):
  """Generate random keys for each leaf in a tree.

  Args:
    rng_key: a JAX pseudo random number generator key.
    tree_like: a nested structure.

  Returns:
    a new key, and a tree of keys with same shape as `tree_like`.
  """
  leaves, treedef = jax.tree_util.tree_flatten(tree_like)
  rng_key, *keys = jax.random.split(rng_key, num=len(leaves) + 1)
  return rng_key, jax.tree_util.tree_unflatten(treedef, keys)


def tree_split_leaves(tree_like: Any,
                      axis: int = 0,
                      keepdim: bool = False):
  """Splits a tree of arrays into an array of trees avoiding data copying.

  Note: `jax.numpy.DeviceArray`'s data gets copied.

  Args:
    tree_like: a nested object with leaves to split.
    axis: an axis for splitting.
    keepdim: a bool indicating whether to keep `axis` dimension.

  Returns:
   A tuple of `size(axis)` trees containing results of splitting.
  """

  # Disable pylint to correctly process `np.ndarray`s.
  if len(tree_like) == 0:  # pylint: disable=g-explicit-length-test
    return tree_like
  leaves, treedef = jax.tree_flatten(tree_like)
  axis_size = leaves[0].shape[axis]
  split_leaves = [np.split(l, axis_size, axis=axis) for l in leaves]
  ind_ = lambda x, i: x[i] if keepdim else np.squeeze(x[i], axis)
  split_trees = ((ind_(l, i) for l in split_leaves) for i in range(axis_size))
  return tuple(jax.tree_unflatten(treedef, t) for t in split_trees)


def tree_replace_masked(tree_data, tree_replacement, mask):
  """Replace slices of the leaves when mask is 1.

  Args:
    tree_data: a nested object with leaves to mask.
    tree_replacement: nested object with the same structure of `tree_data`,
      that cointains the data to insert according to `mask`. If `None`,
      then the masked elements in `tree_data` will be replaced with zeros.
    mask: a mask of 0/1s, whose shape is a prefix of the shape of the leaves
      in `tree_data` and in `tree_replacement`.

  Returns:
    the updated tensor.
  """
  if tree_replacement is None:
    tree_replacement = jax.tree_map(jnp.zeros_like, tree_data)
  return jax.tree_map(
      lambda data, replacement: base.replace_masked(data, replacement, mask),
      tree_data, tree_replacement)


def tree_fn(fn, **unmapped_kwargs):
  """Wrap a function to jax.tree_map over its arguments.

  You may set some named arguments via a partial to skip the `tree_map` on those
  arguments. Usual caveats of `partial` apply (e.g. set named args must be a
  suffix of the argument list).

  Args:
    fn: the function to be wrapped.
    **unmapped_kwargs: the named arguments to be set via a partial.

  Returns:
    a function
  """
  pfn = functools.partial(fn, **unmapped_kwargs)
  def _wrapped(*args):
    return jax.tree_map(pfn, *args)
  return _wrapped


def transpose_last_axis_to_first(tree: Array):
  """Function to transpose the last axis to be first for all leaves in a pytree.

  This function will transpose the last axis to the front for all leaves in a
  pytree; each leaf with shape [D_1, ..., D_{n-1}, D_n] will have shape
  [D_n, D_1, ..., D_{n-1}].

  Args:
    tree: the pytree of Arrays to be transposed.

  Returns:
    tree: the transposed output tree.
  """
  def _transpose(tree):
    return jnp.transpose(tree, [tree.ndim - 1] + list(range(tree.ndim - 1)))
  return tree_fn(_transpose)(tree)


def transpose_first_axis_to_last(tree: Array):
  """Function to transpose the first axis to be last for all leaves in a pytree.

  This function will transpose the first axis to the last dim of all leaves in a
  pytree; each leaf with shape [D_1, D_2, ..., D_n] will have shape
  [D_2, ..., D_n, D_1].

  Args:
    tree: the pytree of Arrays to be transposed.

  Returns:
    tree: the transposed output tree.
  """
  def _transpose(tree):
    return jnp.transpose(tree, list(range(1, tree.ndim)) + [0])
  return tree_fn(_transpose)(tree)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions for working with probability distributions.

Reinforcement learning algorithms often require to estimate suitably
parametrized probability distributions. In this subpackage a distribution is
represented as a collection of functions that may be used to sample from a
distribution, to evaluate its probability mass (or density) function, and
to compute statistics such as its entropy.
"""

import collections
import warnings

import chex
import distrax
import jax.numpy as jnp

Array = chex.Array
Numeric = chex.Numeric


DiscreteDistribution = collections.namedtuple(
    "DiscreteDistribution", ["sample", "probs", "logprob", "entropy", "kl"])
ContinuousDistribution = collections.namedtuple(
    "ContinuousDistribution", ["sample", "prob", "logprob", "entropy",
                               "kl_to_standard_normal", "kl"])


def categorical_sample(key, probs):
  """Sample from a set of discrete probabilities."""
  warnings.warn(
      "Rlax categorical_sample will be deprecated. "
      "Please use distrax.Categorical.sample instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return distrax.Categorical(probs=probs).sample(seed=key)


def softmax(temperature=1.):
  """A softmax distribution."""
  warnings.warn(
      "Rlax softmax will be deprecated. "
      "Please use distrax.Softmax instead.",
      PendingDeprecationWarning, stacklevel=2
  )

  def sample_fn(key: Array, logits: Array):
    return distrax.Softmax(logits, temperature).sample(seed=key)

  def probs_fn(logits: Array):
    return distrax.Softmax(logits, temperature).probs

  def logprob_fn(sample: Array, logits: Array):
    return distrax.Softmax(logits, temperature).log_prob(sample)

  def entropy_fn(logits: Array):
    return distrax.Softmax(logits, temperature).entropy()

  def kl_fn(p_logits: Array, q_logits: Array):
    return categorical_kl_divergence(p_logits, q_logits, temperature)

  return DiscreteDistribution(sample_fn, probs_fn, logprob_fn, entropy_fn,
                              kl_fn)


def clipped_entropy_softmax(temperature=1., entropy_clip=1.):
  """A softmax distribution with clipped entropy (1 is eq to not clipping)."""

  warnings.warn(
      "Rlax clipped_entropy_softmax will be deprecated. "
      "Please use distrax.Softmax instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  def sample_fn(key: Array, logits: Array, action_spec=None):
    del action_spec
    return distrax.Softmax(logits, temperature).sample(seed=key)

  def probs_fn(logits: Array, action_spec=None):
    del action_spec
    return distrax.Softmax(logits, temperature).probs

  def logprob_fn(sample: Array, logits: Array, action_spec=None):
    del action_spec
    return distrax.Softmax(logits, temperature).log_prob(sample)

  def entropy_fn(logits: Array):
    return jnp.minimum(
        distrax.Softmax(logits, temperature).entropy(),
        entropy_clip * jnp.log(logits.shape[-1]))

  def kl_fn(p_logits: Array, q_logits: Array):
    return categorical_kl_divergence(p_logits, q_logits, temperature)

  return DiscreteDistribution(sample_fn, probs_fn, logprob_fn, entropy_fn,
                              kl_fn)


def greedy():
  """A greedy distribution."""
  warnings.warn(
      "Rlax greedy will be deprecated. "
      "Please use distrax.Greedy instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  def sample_fn(key: Array, preferences: Array):
    return distrax.Greedy(preferences).sample(seed=key)

  def probs_fn(preferences: Array):
    return distrax.Greedy(preferences).probs

  def log_prob_fn(sample: Array, preferences: Array):
    return distrax.Greedy(preferences).log_prob(sample)

  def entropy_fn(preferences: Array):
    return distrax.Greedy(preferences).entropy()

  return DiscreteDistribution(sample_fn, probs_fn, log_prob_fn, entropy_fn,
                              None)


def epsilon_greedy(epsilon=None):
  """An epsilon-greedy distribution."""

  warnings.warn(
      "Rlax epsilon_greedy will be deprecated. "
      "Please use distrax.EpsilonGreedy instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  def sample_fn(key: Array, preferences: Array, epsilon=epsilon):
    return distrax.EpsilonGreedy(preferences, epsilon).sample(seed=key)

  def probs_fn(preferences: Array, epsilon=epsilon):
    return distrax.EpsilonGreedy(preferences, epsilon).probs

  def logprob_fn(sample: Array, preferences: Array, epsilon=epsilon):
    return distrax.EpsilonGreedy(preferences, epsilon).log_prob(sample)

  def entropy_fn(preferences: Array, epsilon=epsilon):
    return distrax.EpsilonGreedy(preferences, epsilon).entropy()

  return DiscreteDistribution(sample_fn, probs_fn, logprob_fn, entropy_fn, None)


def gaussian_diagonal(sigma=None):
  """A gaussian distribution with diagonal covariance matrix."""

  warnings.warn(
      "Rlax gaussian_diagonal will be deprecated. "
      "Please use distrax MultivariateNormalDiag instead.",
      PendingDeprecationWarning, stacklevel=2
  )

  def sample_fn(key: Array, mu: Array, sigma: Array = sigma):
    return distrax.MultivariateNormalDiag(
        mu, jnp.ones_like(mu) * sigma).sample(seed=key)

  def prob_fn(sample: Array, mu: Array, sigma: Array = sigma):
    return distrax.MultivariateNormalDiag(
        mu, jnp.ones_like(mu) * sigma).prob(sample)

  def logprob_fn(sample: Array, mu: Array, sigma: Array = sigma):
    return distrax.MultivariateNormalDiag(
        mu, jnp.ones_like(mu) * sigma).log_prob(sample)

  def entropy_fn(mu: Array, sigma: Array = sigma):
    return distrax.MultivariateNormalDiag(
        mu, jnp.ones_like(mu) * sigma).entropy()

  def kl_to_standard_normal_fn(mu: Array, sigma: Array = sigma):
    return distrax.MultivariateNormalDiag(
        mu, jnp.ones_like(mu) * sigma).kl_divergence(
            distrax.MultivariateNormalDiag(
                jnp.zeros_like(mu), jnp.ones_like(mu)))

  def kl_fn(mu_0: Array, sigma_0: Numeric, mu_1: Array, sigma_1: Numeric):
    return distrax.MultivariateNormalDiag(
        mu_0, jnp.ones_like(mu_0) * sigma_0).kl_divergence(
            distrax.MultivariateNormalDiag(mu_1, jnp.ones_like(mu_1) * sigma_1))

  return ContinuousDistribution(sample_fn, prob_fn, logprob_fn, entropy_fn,
                                kl_to_standard_normal_fn, kl_fn)


def squashed_gaussian(sigma_min=-4, sigma_max=0.):
  """A squashed gaussian distribution with diagonal covariance matrix."""

  warnings.warn(
      "Rlax squashed_gaussian will be deprecated. "
      "Please use distrax Transformed MultivariateNormalDiag distribution "
      "with chained Tanh/ScalarAffine bijector instead.",
      PendingDeprecationWarning, stacklevel=2
  )

  def sigma_activation(sigma, sigma_min=sigma_min, sigma_max=sigma_max):
    return jnp.exp(sigma_min + 0.5 * (sigma_max - sigma_min) *
                   (jnp.tanh(sigma) + 1.))

  def mu_activation(mu):
    return jnp.tanh(mu)

  def get_squashed_gaussian_dist(mu, sigma, action_spec=None):
    if action_spec is not None:
      scale = 0.5 * (action_spec.maximum - action_spec.minimum)
      shift = action_spec.minimum
      bijector = distrax.Chain([distrax.ScalarAffine(shift=shift, scale=scale),
                                distrax.ScalarAffine(shift=1.0),
                                distrax.Tanh()])
    else:
      bijector = distrax.Tanh()
    return distrax.Transformed(
        distribution=distrax.MultivariateNormalDiag(
            loc=mu_activation(mu), scale_diag=sigma_activation(sigma)),
        bijector=distrax.Block(bijector, ndims=1))

  def sample_fn(key: Array, mu: Array, sigma: Array, action_spec):
    return get_squashed_gaussian_dist(mu, sigma, action_spec).sample(seed=key)

  def prob_fn(sample: Array, mu: Array, sigma: Array, action_spec):
    return get_squashed_gaussian_dist(mu, sigma, action_spec).prob(sample)

  def logprob_fn(sample: Array, mu: Array, sigma: Array, action_spec):
    return get_squashed_gaussian_dist(mu, sigma, action_spec).log_prob(sample)

  def entropy_fn(mu: Array, sigma: Array):
    return get_squashed_gaussian_dist(mu, sigma).distribution.entropy()

  def kl_to_standard_normal_fn(mu: Array, sigma: Array):
    return get_squashed_gaussian_dist(mu, sigma).distribution.kl_divergence(
        distrax.MultivariateNormalDiag(
            jnp.zeros_like(mu), jnp.ones_like(mu)))

  def kl_fn(mu_0: Array, sigma_0: Numeric, mu_1: Array, sigma_1: Numeric):
    return get_squashed_gaussian_dist(mu_0, sigma_0).distribution.kl_divergence(
        get_squashed_gaussian_dist(mu_1, sigma_1).distribution)

  return ContinuousDistribution(sample_fn, prob_fn, logprob_fn, entropy_fn,
                                kl_to_standard_normal_fn, kl_fn)


def categorical_importance_sampling_ratios(pi_logits_t: Array,
                                           mu_logits_t: Array,
                                           a_t: Array) -> Array:
  """Compute importance sampling ratios from logits.

  Args:
    pi_logits_t: unnormalized logits at time t for the target policy.
    mu_logits_t: unnormalized logits at time t for the behavior policy.
    a_t: actions at time t.

  Returns:
    importance sampling ratios.
  """
  warnings.warn(
      "Rlax categorical_importance_sampling_ratios will be deprecated. "
      "Please use distrax.importance_sampling_ratios instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return distrax.importance_sampling_ratios(distrax.Categorical(
      pi_logits_t), distrax.Categorical(mu_logits_t), a_t)


def categorical_cross_entropy(
    labels: Array,
    logits: Array
) -> Array:
  """Computes the softmax cross entropy between sets of logits and labels.

  See "Deep Learning" by Goodfellow et al.
  (http://www.deeplearningbook.org/contents/prob.html). The computation is
  equivalent to:

                  sum_i (labels_i * log_softmax(logits_i))

  Args:
    labels: a valid probability distribution (non-negative, sum to 1).
    logits: unnormalized log probabilities.

  Returns:
    a scalar loss.
  """
  warnings.warn(
      "Rlax categorical_cross_entropy will be deprecated. "
      "Please use distrax.Categorical.cross_entropy instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return distrax.Categorical(probs=labels).cross_entropy(
      distrax.Categorical(logits=logits))


def categorical_kl_divergence(
    p_logits: Array,
    q_logits: Array,
    temperature: float = 1.
) -> Array:
  """Compute the KL between two categorical distributions from their logits.

  Args:
    p_logits: unnormalized logits for the first distribution.
    q_logits: unnormalized logits for the second distribution.
    temperature: the temperature for the softmax distribution, defaults at 1.

  Returns:
    the kl divergence between the distributions.
  """
  warnings.warn(
      "Rlax categorical_kl_divergence will be deprecated. "
      "Please use distrax.Softmax.kl_divergence instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return distrax.Softmax(p_logits, temperature).kl_divergence(
      distrax.Softmax(q_logits, temperature))


def multivariate_normal_kl_divergence(
    mu_0: Array, sigma_0: Numeric, mu_1: Array, sigma_1: Numeric,
) -> Array:
  """Compute the KL between 2 gaussian distrs with diagonal covariance matrices.

  Args:
    mu_0: array like of mean values for policy 0
    sigma_0: array like of std values for policy 0
    mu_1: array like of mean values for policy 1
    sigma_1: array like of std values for policy 1

  Returns:
    the kl divergence between the distributions.
  """
  warnings.warn(
      "Rlax multivariate_normal_kl_divergence will be deprecated."
      "Please use distrax.MultivariateNormalDiag.kl_divergence instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return distrax.MultivariateNormalDiag(mu_0, sigma_0).kl_divergence(
      distrax.MultivariateNormalDiag(mu_1, sigma_1))



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions for state value and action-value learning.

Value functions estimate the expected return (discounted sum of rewards) that
can be collected by an agent under a given policy of behaviour. This subpackage
implements a number of functions for value learning in discrete scalar action
spaces. Actions are assumed to be represented as indices in the range `[0, A)`
where `A` is the number of distinct actions.
"""
from typing import Union
import chex
import jax
import jax.numpy as jnp
from rlax._src import base
from rlax._src import clipping
from rlax._src import distributions
from rlax._src import multistep

Array = chex.Array
Numeric = chex.Numeric


def td_learning(
    v_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    v_t: Numeric,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the TD-learning temporal difference error.

  See "Learning to Predict by the Methods of Temporal Differences" by Sutton.
  (https://link.springer.com/article/10.1023/A:1022633531479).

  Args:
    v_tm1: state values at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    v_t: state values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
  Returns:
    TD-learning temporal difference error.
  """

  chex.assert_rank([v_tm1, r_t, discount_t, v_t], 0)
  chex.assert_type([v_tm1, r_t, discount_t, v_t], float)

  target_tm1 = r_t + discount_t * v_t
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - v_tm1


def td_lambda(
    v_tm1: Array,
    r_t: Array,
    discount_t: Array,
    v_t: Array,
    lambda_: Numeric,
    stop_target_gradients: bool = True,
) -> Array:
  """Calculates the TD(lambda) temporal difference error.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node74.html).

  Args:
    v_tm1: sequence of state values at time t-1.
    r_t: sequence of rewards at time t.
    discount_t: sequence of discounts at time t.
    v_t: sequence of state values at time t.
    lambda_: mixing parameter lambda, either a scalar or a sequence.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
  Returns:
    TD(lambda) temporal difference error.
  """
  chex.assert_rank([v_tm1, r_t, discount_t, v_t, lambda_], [1, 1, 1, 1, {0, 1}])
  chex.assert_type([v_tm1, r_t, discount_t, v_t, lambda_], float)

  target_tm1 = multistep.lambda_returns(r_t, discount_t, v_t, lambda_)
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - v_tm1


def sarsa(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t: Array,
    a_t: Numeric,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the SARSA temporal difference error.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node64.html.)

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t: Q-values at time t.
    a_t: action index at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.
  Returns:
    SARSA temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t, a_t],
                   [1, 0, 0, 0, 1, 0])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t, a_t],
                   [float, int, float, float, float, int])

  target_tm1 = r_t + discount_t * q_t[a_t]
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def expected_sarsa(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t: Array,
    probs_a_t: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the expected SARSA (SARSE) temporal difference error.

  See "A Theoretical and Empirical Analysis of Expected Sarsa" by Seijen,
  van Hasselt, Whiteson et al.
  (http://www.cs.ox.ac.uk/people/shimon.whiteson/pubs/vanseijenadprl09.pdf).

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t: Q-values at time t.
    probs_a_t: action probabilities at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Expected SARSA temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t, probs_a_t],
                   [1, 0, 0, 0, 1, 1])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t, probs_a_t],
                   [float, int, float, float, float, float])

  target_tm1 = r_t + discount_t * jnp.dot(q_t, probs_a_t)
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def sarsa_lambda(
    q_tm1: Array,
    a_tm1: Array,
    r_t: Array,
    discount_t: Array,
    q_t: Array,
    a_t: Array,
    lambda_: Numeric,
    stop_target_gradients: bool = True,
) -> Array:
  """Calculates the SARSA(lambda) temporal difference error.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node77.html).

  Args:
    q_tm1: sequence of Q-values at time t-1.
    a_tm1: sequence of action indices at time t-1.
    r_t: sequence of rewards at time t.
    discount_t: sequence of discounts at time t.
    q_t: sequence of Q-values at time t.
    a_t: sequence of action indices at time t.
    lambda_: mixing parameter lambda, either a scalar or a sequence.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    SARSA(lambda) temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t, a_t, lambda_],
                   [2, 1, 1, 1, 2, 1, {0, 1}])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t, a_t, lambda_],
                   [float, int, float, float, float, int, float])

  qa_tm1 = base.batched_index(q_tm1, a_tm1)
  qa_t = base.batched_index(q_t, a_t)
  target_tm1 = multistep.lambda_returns(r_t, discount_t, qa_t, lambda_)
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - qa_tm1


def q_learning(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the Q-learning temporal difference error.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node65.html).

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t: Q-values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Q-learning temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t], [1, 0, 0, 0, 1])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t],
                   [float, int, float, float, float])

  target_tm1 = r_t + discount_t * jnp.max(q_t)
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def double_q_learning(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t_value: Array,
    q_t_selector: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the double Q-learning temporal difference error.

  See "Double Q-learning" by van Hasselt.
  (https://papers.nips.cc/paper/3964-double-q-learning.pdf).

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t_value: Q-values at time t.
    q_t_selector: selector Q-values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Double Q-learning temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t_value, q_t_selector],
                   [1, 0, 0, 0, 1, 1])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t_value, q_t_selector],
                   [float, int, float, float, float, float])

  target_tm1 = r_t + discount_t * q_t_value[q_t_selector.argmax()]
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def persistent_q_learning(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t: Array,
    action_gap_scale: float,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the persistent Q-learning temporal difference error.

  See "Increasing the Action Gap: New Operators for Reinforcement Learning"
  by Bellemare, Ostrovski, Guez et al. (https://arxiv.org/abs/1512.04860).

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t: Q-values at time t.
    action_gap_scale: coefficient in [0, 1] for scaling the action gap term.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Persistent Q-learning temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t], [1, 0, 0, 0, 1])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t],
                   [float, int, float, float, float])

  corrected_q_t = (
      (1. - action_gap_scale) * jnp.max(q_t)
      + action_gap_scale * q_t[a_tm1]
  )
  target_tm1 = r_t + discount_t * corrected_q_t
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def qv_learning(
    q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    v_t: Numeric,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the QV-learning temporal difference error.

  See "Two Novel On-policy Reinforcement Learning Algorithms based on
  TD(lambda)-methods" by Wiering and van Hasselt
  (https://ieeexplore.ieee.org/abstract/document/4220845.)

  Args:
    q_tm1: Q-values at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    v_t: state values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    QV-learning temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, v_t], [1, 0, 0, 0, 0])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, v_t],
                   [float, int, float, float, float])

  target_tm1 = r_t + discount_t * v_t
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1[a_tm1]


def qv_max(
    v_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_t: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Calculates the QVMAX temporal difference error.

  See "The QV Family Compared to Other Reinforcement Learning Algorithms" by
  Wiering and van Hasselt (2009).
  (http://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.713.1931)

  Args:
    v_tm1: state values at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_t: Q-values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    QVMAX temporal difference error.
  """
  chex.assert_rank([v_tm1, r_t, discount_t, q_t], [0, 0, 0, 1])
  chex.assert_type([v_tm1, r_t, discount_t, q_t], float)

  target_tm1 = r_t + discount_t * jnp.max(q_t)
  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - v_tm1


def q_lambda(
    q_tm1: Array,
    a_tm1: Array,
    r_t: Array,
    discount_t: Array,
    q_t: Array,
    lambda_: Numeric,
    stop_target_gradients: bool = True,
) -> Array:
  """Calculates Peng's or Watkins' Q(lambda) temporal difference error.

  See "Reinforcement Learning: An Introduction" by Sutton and Barto.
  (http://incompleteideas.net/book/ebook/node78.html).

  Args:
    q_tm1: sequence of Q-values at time t-1.
    a_tm1: sequence of action indices at time t-1.
    r_t: sequence of rewards at time t.
    discount_t: sequence of discounts at time t.
    q_t: sequence of Q-values at time t.
    lambda_: mixing parameter lambda, either a scalar (e.g. Peng's Q(lambda)) or
      a sequence (e.g. Watkin's Q(lambda)).
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Q(lambda) temporal difference error.
  """
  chex.assert_rank([q_tm1, a_tm1, r_t, discount_t, q_t, lambda_],
                   [2, 1, 1, 1, 2, {0, 1}])
  chex.assert_type([q_tm1, a_tm1, r_t, discount_t, q_t, lambda_],
                   [float, int, float, float, float, float])

  qa_tm1 = base.batched_index(q_tm1, a_tm1)
  v_t = jnp.max(q_t, axis=-1)
  target_tm1 = multistep.lambda_returns(r_t, discount_t, v_t, lambda_)

  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - qa_tm1


def retrace(
    q_tm1: Array,
    q_t: Array,
    a_tm1: Array,
    a_t: Array,
    r_t: Array,
    discount_t: Array,
    pi_t: Array,
    mu_t: Array,
    lambda_: float,
    eps: float = 1e-8,
    stop_target_gradients: bool = True,
) -> Array:
  """Calculates Retrace errors.

  See "Safe and Efficient Off-Policy Reinforcement Learning" by Munos et al.
  (https://arxiv.org/abs/1606.02647).

  Args:
    q_tm1: Q-values at time t-1.
    q_t: Q-values at time t.
    a_tm1: action index at time t-1.
    a_t: action index at time t.
    r_t: reward at time t.
    discount_t: discount at time t.
    pi_t: target policy probs at time t.
    mu_t: behavior policy probs at time t.
    lambda_: scalar mixing parameter lambda.
    eps: small value to add to mu_t for numerical stability.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Retrace error.
  """
  chex.assert_rank([q_tm1, q_t, a_tm1, a_t, r_t, discount_t, pi_t, mu_t],
                   [2, 2, 1, 1, 1, 1, 2, 1])
  chex.assert_type([q_tm1, q_t, a_tm1, a_t, r_t, discount_t, pi_t, mu_t],
                   [float, float, int, int, float, float, float, float])

  pi_a_t = base.batched_index(pi_t, a_t)
  c_t = jnp.minimum(1.0, pi_a_t / (mu_t + eps)) * lambda_
  target_tm1 = multistep.general_off_policy_returns_from_action_values(
      q_t, a_t, r_t, discount_t, c_t, pi_t)

  q_a_tm1 = base.batched_index(q_tm1, a_tm1)

  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_a_tm1


def retrace_continuous(q_tm1: Array,
                       q_t: Array,
                       v_t: Array,
                       r_t: Array,
                       discount_t: Array,
                       log_rhos: Array,
                       lambda_: Union[Array, float],
                       stop_target_gradients: bool = True) -> Array:
  """Retrace continuous.

  See "Safe and Efficient Off-Policy Reinforcement Learning" by Munos et al.
  (https://arxiv.org/abs/1606.02647).

  Args:
    q_tm1: Q-values at times [0, ..., K - 1].
    q_t: Q-values evaluated at actions collected using behavior
      policy at times [1, ..., K - 1].
    v_t: Value estimates of the target policy at times [1, ..., K].
    r_t: reward at times [1, ..., K].
    discount_t: discount at times [1, ..., K].
    log_rhos: Log importance weight pi_target/pi_behavior evaluated at actions
      collected using behavior policy [1, ..., K - 1].
    lambda_: scalar or a vector of mixing parameter lambda.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Retrace error.
  """

  chex.assert_rank([q_tm1, q_t, r_t, discount_t, log_rhos, lambda_],
                   [1, 1, 1, 1, 1, {0, 1}])
  chex.assert_type([q_tm1, q_t, r_t, discount_t, log_rhos],
                   [float, float, float, float, float])

  c_t = jnp.minimum(1.0, jnp.exp(log_rhos)) * lambda_

  # The generalized returns are independent of Q-values and cs at the final
  # state.
  target_tm1 = multistep.general_off_policy_returns_from_q_and_v(
      q_t, v_t, r_t, discount_t, c_t)

  target_tm1 = jax.lax.select(stop_target_gradients,
                              jax.lax.stop_gradient(target_tm1), target_tm1)
  return target_tm1 - q_tm1


def categorical_l2_project(
    z_p: Array,
    probs: Array,
    z_q: Array
) -> Array:
  """Projects a categorical distribution (z_p, p) onto a different support z_q.

  The projection step minimizes an L2-metric over the cumulative distribution
  functions (CDFs) of the source and target distributions.

  Let kq be len(z_q) and kp be len(z_p). This projection works for any
  support z_q, in particular kq need not be equal to kp.

  See "A Distributional Perspective on RL" by Bellemare et al.
  (https://arxiv.org/abs/1707.06887).

  Args:
    z_p: support of distribution p.
    probs: probability values.
    z_q: support to project distribution (z_p, probs) onto.

  Returns:
    Projection of (z_p, p) onto support z_q under Cramer distance.
  """
  chex.assert_rank([z_p, probs, z_q], 1)
  chex.assert_type([z_p, probs, z_q], float)

  kp = z_p.shape[0]
  kq = z_q.shape[0]

  # Construct helper arrays from z_q.
  d_pos = jnp.roll(z_q, shift=-1)
  d_neg = jnp.roll(z_q, shift=1)

  # Clip z_p to be in new support range (vmin, vmax).
  z_p = jnp.clip(z_p, z_q[0], z_q[-1])[None, :]
  assert z_p.shape == (1, kp)

  # Get the distance between atom values in support.
  d_pos = (d_pos - z_q)[:, None]  # z_q[i+1] - z_q[i]
  d_neg = (z_q - d_neg)[:, None]  # z_q[i] - z_q[i-1]
  z_q = z_q[:, None]
  assert z_q.shape == (kq, 1)

  # Ensure that we do not divide by zero, in case of atoms of identical value.
  d_neg = jnp.where(d_neg > 0, 1. / d_neg, jnp.zeros_like(d_neg))
  d_pos = jnp.where(d_pos > 0, 1. / d_pos, jnp.zeros_like(d_pos))

  delta_qp = z_p - z_q  # clip(z_p)[j] - z_q[i]
  d_sign = (delta_qp >= 0.).astype(probs.dtype)
  assert delta_qp.shape == (kq, kp)
  assert d_sign.shape == (kq, kp)

  # Matrix of entries sgn(a_ij) * |a_ij|, with a_ij = clip(z_p)[j] - z_q[i].
  delta_hat = (d_sign * delta_qp * d_pos) - ((1. - d_sign) * delta_qp * d_neg)
  probs = probs[None, :]
  assert delta_hat.shape == (kq, kp)
  assert probs.shape == (1, kp)

  return jnp.sum(jnp.clip(1. - delta_hat, 0., 1.) * probs, axis=-1)


def categorical_td_learning(
    v_atoms_tm1: Array,
    v_logits_tm1: Array,
    r_t: Numeric,
    discount_t: Numeric,
    v_atoms_t: Array,
    v_logits_t: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Implements TD-learning for categorical value distributions.

  See "A Distributional Perspective on Reinforcement Learning", by
  Bellemere, Dabney and Munos (https://arxiv.org/pdf/1707.06887.pdf).

  Args:
    v_atoms_tm1: atoms of V distribution at time t-1.
    v_logits_tm1: logits of V distribution at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    v_atoms_t: atoms of V distribution at time t.
    v_logits_t: logits of V distribution at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Categorical Q learning loss (i.e. temporal difference error).
  """
  chex.assert_rank(
      [v_atoms_tm1, v_logits_tm1, r_t, discount_t, v_atoms_t, v_logits_t],
      [1, 1, 0, 0, 1, 1])
  chex.assert_type(
      [v_atoms_tm1, v_logits_tm1, r_t, discount_t, v_atoms_t, v_logits_t],
      [float, float, float, float, float, float])

  # Scale and shift time-t distribution atoms by discount and reward.
  target_z = r_t + discount_t * v_atoms_t

  # Convert logits to distribution.
  v_t_probs = jax.nn.softmax(v_logits_t)

  # Project using the Cramer distance and maybe stop gradient flow to targets.
  target = categorical_l2_project(target_z, v_t_probs, v_atoms_tm1)
  target = jax.lax.select(stop_target_gradients, jax.lax.stop_gradient(target),
                          target)

  # Compute loss (i.e. temporal difference error).
  return distributions.categorical_cross_entropy(
      labels=target, logits=v_logits_tm1)


def categorical_q_learning(
    q_atoms_tm1: Array,
    q_logits_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_atoms_t: Array,
    q_logits_t: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Implements Q-learning for categorical Q distributions.

  See "A Distributional Perspective on Reinforcement Learning", by
  Bellemere, Dabney and Munos (https://arxiv.org/pdf/1707.06887.pdf).

  Args:
    q_atoms_tm1: atoms of Q distribution at time t-1.
    q_logits_tm1: logits of Q distribution at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_atoms_t: atoms of Q distribution at time t.
    q_logits_t: logits of Q distribution at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Categorical Q-learning loss (i.e. temporal difference error).
  """
  chex.assert_rank([
      q_atoms_tm1, q_logits_tm1, a_tm1, r_t, discount_t, q_atoms_t, q_logits_t
  ], [1, 2, 0, 0, 0, 1, 2])
  chex.assert_type([
      q_atoms_tm1, q_logits_tm1, a_tm1, r_t, discount_t, q_atoms_t, q_logits_t
  ], [float, float, int, float, float, float, float])

  # Scale and shift time-t distribution atoms by discount and reward.
  target_z = r_t + discount_t * q_atoms_t

  # Convert logits to distribution, then find greedy action in state s_t.
  q_t_probs = jax.nn.softmax(q_logits_t)
  q_t_mean = jnp.sum(q_t_probs * q_atoms_t[jnp.newaxis, :], axis=1)
  pi_t = jnp.argmax(q_t_mean)

  # Compute distribution for greedy action.
  p_target_z = q_t_probs[pi_t]

  # Project using the Cramer distance and maybe stop gradient flow to targets.
  target = categorical_l2_project(target_z, p_target_z, q_atoms_tm1)
  target = jax.lax.select(stop_target_gradients, jax.lax.stop_gradient(target),
                          target)

  # Compute loss (i.e. temporal difference error).
  logit_qa_tm1 = q_logits_tm1[a_tm1]
  return distributions.categorical_cross_entropy(
      labels=target, logits=logit_qa_tm1)


def categorical_double_q_learning(
    q_atoms_tm1: Array,
    q_logits_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    q_atoms_t: Array,
    q_logits_t: Array,
    q_t_selector: Array,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Implements double Q-learning for categorical Q distributions.

  See "A Distributional Perspective on Reinforcement Learning", by
  Bellemere, Dabney and Munos (https://arxiv.org/pdf/1707.06887.pdf)
  and "Double Q-learning" by van Hasselt.
  (https://papers.nips.cc/paper/3964-double-q-learning.pdf).

  Args:
    q_atoms_tm1: atoms of Q distribution at time t-1.
    q_logits_tm1: logits of Q distribution at time t-1.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    q_atoms_t: atoms of Q distribution at time t.
    q_logits_t: logits of Q distribution at time t.
    q_t_selector: selector Q-values at time t.
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Categorical double Q-learning loss (i.e. temporal difference error).
  """
  chex.assert_rank([
      q_atoms_tm1, q_logits_tm1, a_tm1, r_t, discount_t, q_atoms_t, q_logits_t,
      q_t_selector
  ], [1, 2, 0, 0, 0, 1, 2, 1])
  chex.assert_type([
      q_atoms_tm1, q_logits_tm1, a_tm1, r_t, discount_t, q_atoms_t, q_logits_t,
      q_t_selector
  ], [float, float, int, float, float, float, float, float])

  # Scale and shift time-t distribution atoms by discount and reward.
  target_z = r_t + discount_t * q_atoms_t

  # Select logits for greedy action in state s_t and convert to distribution.
  p_target_z = jax.nn.softmax(q_logits_t[q_t_selector.argmax()])

  # Project using the Cramer distance and maybe stop gradient flow to targets.
  target = categorical_l2_project(target_z, p_target_z, q_atoms_tm1)
  target = jax.lax.select(stop_target_gradients, jax.lax.stop_gradient(target),
                          target)

  # Compute loss (i.e. temporal difference error).
  logit_qa_tm1 = q_logits_tm1[a_tm1]
  return distributions.categorical_cross_entropy(
      labels=target, logits=logit_qa_tm1)


def quantile_regression_loss(
    dist_src: Array,
    tau_src: Array,
    dist_target: Array,
    huber_param: float = 0.,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Compute (Huber) QR loss between two discrete quantile-valued distributions.

  See "Distributional Reinforcement Learning with Quantile Regression" by
  Dabney et al. (https://arxiv.org/abs/1710.10044).

  Args:
    dist_src: source probability distribution.
    tau_src: source distribution probability thresholds.
    dist_target: target probability distribution.
    huber_param: Huber loss parameter, defaults to 0 (no Huber loss).
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Quantile regression loss.
  """
  chex.assert_rank([dist_src, tau_src, dist_target], 1)
  chex.assert_type([dist_src, tau_src, dist_target], float)

  # Calculate quantile error.
  delta = dist_target[None, :] - dist_src[:, None]
  delta_neg = (delta < 0.).astype(jnp.float32)
  delta_neg = jax.lax.select(stop_target_gradients,
                             jax.lax.stop_gradient(delta_neg), delta_neg)
  weight = jnp.abs(tau_src[:, None] - delta_neg)

  # Calculate Huber loss.
  if huber_param > 0.:
    loss = clipping.huber_loss(delta, huber_param)
  else:
    loss = jnp.abs(delta)
  loss *= weight

  # Average over target-samples dimension, sum over src-samples dimension.
  return jnp.sum(jnp.mean(loss, axis=-1))


def quantile_q_learning(
    dist_q_tm1: Array,
    tau_q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    dist_q_t_selector: Array,
    dist_q_t: Array,
    huber_param: float = 0.,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Implements Q-learning for quantile-valued Q distributions.

  See "Distributional Reinforcement Learning with Quantile Regression" by
  Dabney et al. (https://arxiv.org/abs/1710.10044).

  Args:
    dist_q_tm1: Q distribution at time t-1.
    tau_q_tm1: Q distribution probability thresholds.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    dist_q_t_selector: Q distribution at time t for selecting greedy action in
      target policy. This is separate from dist_q_t as in Double Q-Learning, but
      can be computed with the target network and a separate set of samples.
    dist_q_t: target Q distribution at time t.
    huber_param: Huber loss parameter, defaults to 0 (no Huber loss).
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Quantile regression Q learning loss.
  """
  chex.assert_rank([
      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t_selector, dist_q_t
  ], [2, 1, 0, 0, 0, 2, 2])
  chex.assert_type([
      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t_selector, dist_q_t
  ], [float, float, int, float, float, float, float])

  # Only update the taken actions.
  dist_qa_tm1 = dist_q_tm1[:, a_tm1]

  # Select target action according to greedy policy w.r.t. dist_q_t_selector.
  q_t_selector = jnp.mean(dist_q_t_selector, axis=0)
  a_t = jnp.argmax(q_t_selector)
  dist_qa_t = dist_q_t[:, a_t]

  # Compute target, do not backpropagate into it.
  dist_target = r_t + discount_t * dist_qa_t
  dist_target = jax.lax.select(stop_target_gradients,
                               jax.lax.stop_gradient(dist_target), dist_target)

  return quantile_regression_loss(
      dist_qa_tm1, tau_q_tm1, dist_target, huber_param)


def quantile_expected_sarsa(
    dist_q_tm1: Array,
    tau_q_tm1: Array,
    a_tm1: Numeric,
    r_t: Numeric,
    discount_t: Numeric,
    dist_q_t: Array,
    probs_a_t: Array,
    huber_param: float = 0.,
    stop_target_gradients: bool = True,
) -> Numeric:
  """Implements Expected SARSA for quantile-valued Q distributions.

  Args:
    dist_q_tm1: Q distribution at time t-1.
    tau_q_tm1: Q distribution probability thresholds.
    a_tm1: action index at time t-1.
    r_t: reward at time t.
    discount_t: discount at time t.
    dist_q_t: target Q distribution at time t.
    probs_a_t: action probabilities at time t.
    huber_param: Huber loss parameter, defaults to 0 (no Huber loss).
    stop_target_gradients: bool indicating whether or not to apply stop gradient
      to targets.

  Returns:
    Quantile regression Expected SARSA learning loss.
  """
  chex.assert_rank([
      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t, probs_a_t
  ], [2, 1, 0, 0, 0, 2, 1])
  chex.assert_type([
      dist_q_tm1, tau_q_tm1, a_tm1, r_t, discount_t, dist_q_t, probs_a_t
  ], [float, float, int, float, float, float, float])

  # Only update the taken actions.
  dist_qa_tm1 = dist_q_tm1[:, a_tm1]

  # Compute target, do not backpropagate into it.
  dist_target = r_t + discount_t * dist_q_t
  dist_target = jax.lax.select(stop_target_gradients,
                               jax.lax.stop_gradient(dist_target), dist_target)
  probs_a_t = jax.lax.select(stop_target_gradients,
                             jax.lax.stop_gradient(probs_a_t), probs_a_t)

  per_action_qr = jax.vmap(
      quantile_regression_loss, in_axes=(None, None, 1, None))
  per_action_loss = per_action_qr(
      dist_qa_tm1, tau_q_tm1, dist_target, huber_param)

  return jnp.dot(per_action_loss, probs_a_t)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for model_learning.py."""

from absl.testing import absltest
import jax.numpy as jnp
import numpy as np
from rlax._src import model_learning


class ModelLearningTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.trajectories = jnp.array([  # [T, B]
        [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
        [0, 10, 20, 30, 40, 50, 60, 70, 80, 90]
    ]).transpose()
    self.start_indices = jnp.array([  # [B, num_starts]
        [0, 1, 4],
        [1, 2, 5]
    ])
    self.invalid_start_indices = jnp.array([  # [B, num_starts]
        [0, 1, 25],  # 25 is out of bound
        [1, 2, 5]])

  def test_extract_subsequences(self):
    output = model_learning.extract_subsequences(
        self.trajectories, self.start_indices, 3)
    expected_output = jnp.array([
        [[0, 1, 4],
         [10, 20, 50]],
        [[1, 2, 5],
         [20, 30, 60]],
        [[2, 3, 6],
         [30, 40, 70]]])
    # Test output.
    np.testing.assert_allclose(output, expected_output)

  def test_extract_subsequences_with_validation_bounds(self):
    with self.assertRaisesRegex(AssertionError, 'Expected len >='):
      model_learning.extract_subsequences(
          self.trajectories, self.invalid_start_indices, 1,
          max_valid_start_idx=24)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions to construct and learn generalized value functions.

According to the reward hypothesis (Sutton et al., 2018) any goal might be
formulated as a suitable scalar `cumulant` to be maximized. Generalized value
functions (Sutton et al. 2011) extend the notion of value functions to include
estimates of discounted sums of `cumulants` different from the main task reward.
"""

import chex
import jax.numpy as jnp

Array = chex.Array


def pixel_control_rewards(
    observations: Array,
    cell_size: int,
) -> Array:
  """Calculates cumulants for pixel control tasks from an observation sequence.

  The observations are first split in a grid of KxK cells. For each cell a
  distinct pseudo reward is computed as the average absolute change in pixel
  intensity across all pixels in the cell. The change in intensity is averaged
  across both pixels and channels (e.g. RGB).

  The `observations` provided to this function should be cropped suitably, to
  ensure that the observations' height and width are a multiple of `cell_size`.
  The values of the `observations` tensor should be rescaled to [0, 1].

  See "Reinforcement Learning with Unsupervised Auxiliary Tasks" by Jaderberg,
  Mnih, Czarnecki et al. (https://arxiv.org/abs/1611.05397).

  Args:
    observations: A tensor of shape `[T+1,H,W,C]`, where
      * `T` is the sequence length,
      * `H` is height,
      * `W` is width,
      * `C` is a channel dimension.
    cell_size: The size of each cell.

  Returns:
    A tensor of pixel control rewards calculated from the observation. The
    shape is `[T,H',W']`, where `H'=H/cell_size` and `W'=W/cell_size`.
  """
  chex.assert_rank(observations, 4)
  chex.assert_type(observations, float)

  # Shape info.
  h = observations.shape[1] // cell_size  # new height.
  w = observations.shape[2] // cell_size  # new width.
  # Calculate the absolute differences across the sequence.
  abs_diff = jnp.abs(observations[1:] - observations[:-1])
  # Average within cells to get the cumulants.
  abs_diff = abs_diff.reshape(
      (-1, h, cell_size, w, cell_size, observations.shape[3]))
  return abs_diff.mean(axis=(2, 4, 5))


def feature_control_rewards(
    features: Array,
    cumulant_type='absolute_change',
    discount=None,
) -> Array:
  """Calculates cumulants for feature control tasks from a sequence of features.

  For each feature dimension, a distinct pseudo reward is computed based on the
  change in the feature value between consecutive timesteps. Depending on
  `cumulant_type`, cumulants may be equal the features themselves, the absolute
  difference between their values in consecutive steps, their increase/decrease,
  or may take the form of a potential-based reward discounted by `discount`.

  See "Reinforcement Learning with Unsupervised Auxiliary Tasks" by Jaderberg,
  Mnih, Czarnecki et al. (https://arxiv.org/abs/1611.05397).

  Args:
    features: A tensor of shape `[T+1,D]` of features.
    cumulant_type: either 'feature' (feature is the reward), `absolute_change`
      (the reward equals the absolute difference between consecutive
      timesteps), `increase` (the reward equals the increase in the
      value of the feature), `decrease` (the reward equals the decrease in the
      value of the feature), or 'potential' (r=gamma*phi_{t+1} - phi_t).
    discount: (optional) discount for potential based rewards.

  Returns:
    A tensor of cumulants calculated from the features. The shape is `[T,D]`.
  """
  chex.assert_rank(features, 2)
  chex.assert_type(features, float)

  if cumulant_type == 'feature':
    return features[1:]
  elif cumulant_type == 'absolute_change':
    return jnp.abs(features[1:] - features[:-1])
  elif cumulant_type == 'increase':
    return features[1:] - features[:-1]
  elif cumulant_type == 'decrease':
    return features[:-1] - features[1:]
  elif cumulant_type == 'potential':
    return discount * features[1:] - features[:-1]
  else:
    raise ValueError(f'Unknown cumulant_type {cumulant_type}')



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""JAX functions for implementing forms of gradient clipping.

Gradient clipping is commonly used to avoid taking too large steps in parameter
space when updating the parameters of an agent's policy, value or model. Certain
forms of gradient clipping can be conveniently expressed as transformations of
the loss function optimized by a suitable gradient descent algorithm.
"""

import chex
import jax
import jax.numpy as jnp
from jax.tree_util import tree_map

Array = chex.Array


def huber_loss(x: Array, delta: float = 1.) -> Array:
  """Huber loss, similar to L2 loss close to zero, L1 loss away from zero.

  See "Robust Estimation of a Location Parameter" by Huber.
  (https://projecteuclid.org/download/pdf_1/euclid.aoms/1177703732).

  Args:
    x: a vector of arbitrary shape.
    delta: the bounds for the huber loss transformation, defaults at 1.

  Note `grad(huber_loss(x))` is equivalent to `grad(0.5 * clip_gradient(x)**2)`.

  Returns:
    a vector of same shape of `x`.
  """
  chex.assert_type(x, float)

  # 0.5 * x^2                  if |x| <= d
  # 0.5 * d^2 + d * (|x| - d)  if |x| > d
  abs_x = jnp.abs(x)
  quadratic = jnp.minimum(abs_x, delta)
  # Same as max(abs_x - delta, 0) but avoids potentially doubling gradient.
  linear = abs_x - quadratic
  return 0.5 * quadratic**2 + delta * linear


@jax.custom_gradient
def clip_gradient(x, gradient_min: float, gradient_max: float):
  """Identity but the gradient in the backward pass is clipped.

  See "Human-level control through deep reinforcement learning" by Mnih et al,
  (https://www.nature.com/articles/nature14236)

  Note `grad(0.5 * clip_gradient(x)**2)` is equivalent to `grad(huber_loss(x))`.

  Note: x cannot be properly annotated because pytype does not support recursive
  types; we would otherwise use the chex.ArrayTree pytype annotation here. Most
  often x will be a single array of arbitrary shape, but the implementation
  supports pytrees.

  Args:
    x: a pytree of arbitrary shape.
    gradient_min: min elementwise size of the gradient.
    gradient_max: max elementwise size of the gradient.

  Returns:
    a vector of same shape of `x`.
  """
  chex.assert_type(x, float)

  def _compute_gradient(g):
    return (tree_map(lambda g: jnp.clip(g, gradient_min, gradient_max),
                     g), 0., 0.)

  return x, _compute_gradient



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for embedding.py."""

from absl.testing import absltest
from absl.testing import parameterized
import jax
import numpy as np
from rlax._src import embedding


class EmbeddingTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._features = np.array([[1., 2.], [3., 2.]])
    self._num_actions = 3
    self._actions = np.array([1, 2])
    self._rewards = np.array([1., 1])

  def test_embed_zeros(self):
    # Embedding zero feature array.
    zero_features = np.zeros_like(self._features)
    emb = embedding.embed_oar(zero_features, self._actions, self._rewards,
                              self._num_actions)
    np.testing.assert_array_equal(emb[:, :self._features.shape[-1]],
                                  zero_features)

  def test_embed_shape(self):
    # Test output shape [T?, B, D+A+1].
    emb = embedding.embed_oar(self._features, self._actions, self._rewards,
                              self._num_actions)
    np.testing.assert_array_equal(
        emb.shape[-1], self._features.shape[-1] + self._num_actions + 1)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `nonlinear_bellman.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import numpy as np
from rlax._src import nonlinear_bellman


class IdentityTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.q_t = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2]]],
        dtype=np.float64)

  @parameterized.parameters(
      nonlinear_bellman.IDENTITY_PAIR,
      nonlinear_bellman.SIGNED_LOGP1_PAIR,
      nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR,
      nonlinear_bellman.HYPERBOLIC_SIN_PAIR)
  def test_identity(self, tx, inv_tx):
    """Tests that tx(inv_tx(x)) == inv_tx(tx(x)) == x."""
    np.testing.assert_allclose(inv_tx(tx(self.q_t)), self.q_t, rtol=1e-3)
    np.testing.assert_allclose(tx(inv_tx(self.q_t)), self.q_t, rtol=1e-3)

  def test_muzero_pair_is_consistent(self):
    a = np.array([1.03, 4.43, -3012.33, 0.0])
    tx = nonlinear_bellman.muzero_pair(
        num_bins=601,
        min_value=-300,
        max_value=300,
        tx=nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR)
    probs = tx.apply(a)
    scalar = tx.apply_inv(probs)
    np.testing.assert_allclose(a, scalar, rtol=1e-4, atol=1e-4)

  def test_unbiased_transform_pair_is_consistent(self):
    a = np.array([1.03, 4.43, -3012.33, 0.0])
    tx = nonlinear_bellman.unbiased_transform_pair(
        num_bins=601,
        min_value=-300,
        max_value=300,
        tx=nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR)
    probs = tx.apply(a)
    scalar = tx.apply_inv(probs)
    np.testing.assert_allclose(a, scalar, rtol=1e-4, atol=1e-4)


class TransformedQLambdaTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.lambda_ = 0.75

    self.q_tm1 = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1]],
         [[2.1, 3.1], [-1.1, 0.1], [-2.1, -1.1]]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [[0, 1, 0],
         [1, 0, 0]],
        dtype=np.int32)
    self.discount_t = np.array(
        [[0., 0.89, 0.85],
         [0.88, 1., 0.83]],
        dtype=np.float32)
    self.r_t = np.array(
        [[-1.3, -1.3, 2.3],
         [1.3, 5.3, -3.3]],
        dtype=np.float32)
    self.q_t = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2]]],
        dtype=np.float32)

    self.expected_td = np.array(
        [[[-2.4, 0.4280, 1.07], [0.6935, 3.478, -2.196]],
         [[-1.9329, 0.6643, -0.7854], [-0.20713, 2.1855, 0.27132]],
         [[-1.6179, 0.4633, -0.7576], [-1.1097, 1.6509, 0.3598]],
         [[-2.1785, 0.6562, -0.5938], [-0.0892, 2.6553, -0.1208]]],
        dtype=np.float32)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('identity0', nonlinear_bellman.IDENTITY_PAIR, 0),
      ('signed_logp11', nonlinear_bellman.SIGNED_LOGP1_PAIR, 1),
      ('signed_hyperbolic2', nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR, 2),
      ('hyperbolic_sin3', nonlinear_bellman.HYPERBOLIC_SIN_PAIR, 3))
  def test_transformed_q_lambda_batch(self, tx_pair, td_index):
    """Tests correctness for full batch."""
    transformed_q_lambda = self.variant(jax.vmap(functools.partial(
        nonlinear_bellman.transformed_q_lambda, tx_pair=tx_pair,
        lambda_=self.lambda_)))
    # Compute vtrace output.
    actual_td = transformed_q_lambda(
        self.q_tm1, self.a_tm1, self.r_t, self.discount_t, self.q_t)
    # Test output.
    np.testing.assert_allclose(self.expected_td[td_index], actual_td, rtol=1e-3)


class TransformedNStepQLearningTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.n = 2

    self.q_tm1 = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1]],
         [[2.1, 3.1], [-1.1, 0.1], [-2.1, -1.1]]],
        dtype=np.float32)
    self.a_tm1 = np.array(
        [[0, 1, 0],
         [1, 0, 0]],
        dtype=np.int32)
    self.discount_t = np.array(
        [[0., 0.89, 0.85],
         [0.88, 1., 0.83]],
        dtype=np.float32)
    self.r_t = np.array(
        [[-1.3, -1.3, 2.3],
         [1.3, 5.3, -3.3]],
        dtype=np.float32)
    self.target_q_t = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2]]],
        dtype=np.float32)
    self.a_t = np.array(
        [[0, 1, 0],
         [1, 0, 0]],
        dtype=np.int32)

    self.expected_td = np.array([
        [[-2.4, 1.3112999, 1.0700002],
         [3.9199996, 2.104, -2.196]],
        [[-1.9329091, 0.9564189, -0.7853615],
         [-0.9021418, 1.1716722, 0.2713145]],
        [[-1.6178751, 0.85600746, -0.75762916],
         [-0.87689304, 0.6246443, 0.3598088]],
        [[-2.178451, 1.02313, -0.593768],
         [-0.415362, 1.790864, -0.120749]]
    ], dtype=np.float32)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('identity0', nonlinear_bellman.IDENTITY_PAIR, 0),
      ('signed_logp11', nonlinear_bellman.SIGNED_LOGP1_PAIR, 1),
      ('signed_hyperbolic2', nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR, 2),
      ('hyperbolic_sin3', nonlinear_bellman.HYPERBOLIC_SIN_PAIR, 3))
  def test_transformed_q_lambda_batch(self, tx_pair, td_index):
    """Tests correctness for full batch."""
    transformed_n_step_q_learning = self.variant(jax.vmap(functools.partial(
        nonlinear_bellman.transformed_n_step_q_learning, tx_pair=tx_pair,
        n=self.n)))
    actual_td = transformed_n_step_q_learning(
        self.q_tm1, self.a_tm1, self.target_q_t, self.a_t, self.r_t,
        self.discount_t)
    np.testing.assert_allclose(self.expected_td[td_index], actual_td, rtol=1e-3)


class TransformedRetraceTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self._lambda = 0.9

    self._qs = np.array(
        [[[1.1, 2.1], [-1.1, 1.1], [3.1, -3.1], [-1.2, 0.0]],
         [[2.1, 3.1], [9.5, 0.1], [-2.1, -1.1], [0.1, 7.4]]],
        dtype=np.float32)
    self._targnet_qs = np.array(
        [[[1.2, 2.2], [-1.2, 0.2], [2.2, -1.2], [-2.25, -6.0]],
         [[4.2, 2.2], [1.2, 1.2], [-1.2, -2.2], [1.5, 1.0]]],
        dtype=np.float32)
    self._actions = np.array(
        [[0, 1, 0, 0],
         [1, 0, 0, 1]],
        dtype=np.int32)
    self._rewards = np.array(
        [[-1.3, -1.3, 2.3, 42.0],
         [1.3, 5.3, -3.3, -5.0]],
        dtype=np.float32)
    self._pcontinues = np.array(
        [[0., 0.89, 0.85, 0.99],
         [0.88, 1., 0.83, 0.95]],
        dtype=np.float32)
    self._target_policy_probs = np.array(
        [[[0.5, 0.5], [0.2, 0.8], [0.6, 0.4], [0.9, 0.1]],
         [[0.1, 0.9], [1.0, 0.0], [0.3, 0.7], [0.7, 0.3]]],
        dtype=np.float32)
    self._behavior_policy_probs = np.array(
        [[0.5, 0.1, 0.9, 0.3],
         [0.4, 0.6, 1.0, 0.9]],
        dtype=np.float32)

    self.expected_td = np.array(
        [[[-2.4, -2.7905, -3.0313], [0.7889, -6.3645, -0.0795]],
         [[-1.9329, -4.2626, -6.7738], [-2.3989, -9.9802, 1.4852]],
         [[-1.6179, -3.0165, -5.2699], [-2.7742, -9.9544, 2.3167]],
         [[-2.1785, -4.2530, -6.7081], [-1.3654, -8.2213, 0.7641]]],
        dtype=np.float32)

  @chex.all_variants()
  @parameterized.named_parameters(
      ('identity0', nonlinear_bellman.IDENTITY_PAIR, 0),
      ('signed_logp11', nonlinear_bellman.SIGNED_LOGP1_PAIR, 1),
      ('signed_hyperbolic2', nonlinear_bellman.SIGNED_HYPERBOLIC_PAIR, 2),
      ('hyperbolic_sin3', nonlinear_bellman.HYPERBOLIC_SIN_PAIR, 3))
  def test_transformed_retrace_batch(self, tx_pair, td_index):
    """Tests correctness for full batch."""
    transformed_retrace = self.variant(jax.vmap(functools.partial(
        nonlinear_bellman.transformed_retrace,
        tx_pair=tx_pair, lambda_=self._lambda)))
    # Compute transformed vtrace td errors in batch.
    actual_td = transformed_retrace(
        self._qs[:, :-1], self._targnet_qs[:, 1:], self._actions[:, :-1],
        self._actions[:, 1:], self._rewards[:, :-1], self._pcontinues[:, :-1],
        self._target_policy_probs[:, 1:], self._behavior_policy_probs[:, 1:])
    # Test output.
    np.testing.assert_allclose(self.expected_td[td_index], actual_td, rtol=1e-3)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for target network switching."""

from typing import Any
import warnings

import chex
import jax

Numeric = chex.Numeric


def conditional_update(new_tensors: Any, old_tensors: Any, is_time: Numeric):
  """Checks whether to update the params and returns the correct params."""
  warnings.warn(
      "Rlax conditional_update will be deprecated. Please use optax instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return jax.tree_map(
      lambda new, old: jax.lax.select(is_time, new, old),
      new_tensors, old_tensors)


def periodic_update(
    new_tensors: Any, old_tensors: Any,
    steps: chex.Array, update_period: int):
  """Periodically switch all elements from a nested struct with new elements."""
  warnings.warn(
      "Rlax periodic_update will be deprecated. Please use optax instead.",
      PendingDeprecationWarning, stacklevel=2
  )
  return conditional_update(
      new_tensors, old_tensors, is_time=steps % update_period == 0)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for losses.py."""

from absl.testing import absltest
import dm_env
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import interruptions


class InterruptionsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.correct_trajectory = jnp.array([
        [dm_env.StepType.MID],
        [dm_env.StepType.MID],
        [dm_env.StepType.MID],
        [dm_env.StepType.LAST],
        [dm_env.StepType.FIRST],
        [dm_env.StepType.MID],
    ])
    self.broken_trajectory = jnp.array([
        [dm_env.StepType.MID],
        [dm_env.StepType.MID],
        [dm_env.StepType.MID],
        [dm_env.StepType.MID],
        [dm_env.StepType.FIRST],
        [dm_env.StepType.MID],
    ])

  def test_fix_step_type_on_interruptions(self):
    output1 = interruptions.fix_step_type_on_interruptions(
        self.correct_trajectory)
    output2 = interruptions.fix_step_type_on_interruptions(
        self.broken_trajectory)
    # Test output.
    np.testing.assert_allclose(output1, self.correct_trajectory)
    np.testing.assert_allclose(output2, self.correct_trajectory)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for pop_art.py."""

import functools

from absl.testing import absltest
from absl.testing import parameterized

import chex
import haiku as hk
from haiku import data_structures
from haiku import initializers
import jax
import jax.nn
import jax.numpy as jnp
import numpy as np
from rlax._src import pop_art

_INPUT_DIM = 3


def setUpModule():
  chex.set_n_cpu_devices(n=4)
  chex.assert_devices_available(n=4, devtype='cpu', backend='cpu')


def get_constant_linear_params(num_outputs):
  w = np.ones([_INPUT_DIM, num_outputs])
  b = np.zeros([num_outputs])
  return dict(w=w, b=b)


def get_fake_pop_art_state(num_outputs):
  """Returns a fake PopArtState."""
  shift = np.arange(num_outputs).astype(np.float32) + 1
  scale = shift
  second_moment = np.square(scale) + np.square(shift)
  return pop_art.PopArtState(shift, scale, second_moment)


class PopArtTest(parameterized.TestCase):

  @parameterized.parameters(
      (None, np.array([[[0., 0., 1.],
                        [0., 5., 0.]],
                       [[9., 0., 0.],
                        [0., 0., 0.]]])),
      ([], np.array([[[0., 0., 1.],
                      [0., 5., 0.]],
                     [[9., 0., 0.],
                      [0., 0., 0.]]])),
      (['i'], np.array([[[0., 5., 1.],
                         [0., 5., 1.]],
                        [[9., 0., 0.],
                         [9., 0., 0.]]])),
      (['j'], np.array([[[9., 0., 1.],
                         [0., 5., 0.]],
                        [[9., 0., 1.],
                         [0., 5., 0.]]])),
      (['i', 'j'], np.array([[[9., 5., 1.],
                              [9., 5., 1.]],
                             [[9., 5., 1.],
                              [9., 5., 1.]]])))
  def test_cross_replica_scatter_add(self, axes, expected):
    shape = (2, 2, 3)
    source = np.zeros(shape)
    fn = functools.partial(pop_art._cross_replica_scatter_add, axis_name=axes)
    mapped_fn = jax.pmap(fn, axis_name='i', backend='cpu')
    mapped_fn = jax.pmap(mapped_fn, axis_name='j', backend='cpu')

    updates = np.array([[[1., 0., 0.],
                         [0., 5., 0.]],
                        [[0., 0., 9.],
                         [0., 0., 0.]]])
    indices = np.array([[[2, 2, 2],
                         [1, 1, 1]],
                        [[0, 0, 0],
                         [0, 1, 2]]])
    np.testing.assert_equal(mapped_fn(source, indices, updates), expected)

  def test_normalize_unnormalize_is_identity(self):
    num_outputs = 3
    state = get_fake_pop_art_state(num_outputs)
    np.random.seed(42)
    # Draws random numbers with the same magnitude as the pop art shift and
    # scales to ensure numerical stability.
    values = np.random.randint(-10, 10, size=(100,)).astype(np.float32)
    indices = np.random.randint(0, num_outputs, size=(100,))
    np.testing.assert_allclose(
        values,
        pop_art.normalize(state, pop_art.unnormalize(state, values, indices),
                          indices))
    np.testing.assert_allclose(
        values,
        pop_art.unnormalize(state, pop_art.normalize(state, values, indices),
                            indices))

  def test_unnormalize_linear(self):
    num_outputs = 3
    state = get_fake_pop_art_state(num_outputs)

    # Verify that it denormalizes as planned.
    inputs = np.ones([num_outputs, num_outputs])
    indices = np.arange(num_outputs)[::-1]
    out = pop_art.unnormalize_linear(state, inputs, indices)

    expected_normalized = np.asarray([1, 1, 1])
    expected_unnormalized = np.asarray([6, 4, 2])
    np.testing.assert_allclose(out.normalized, expected_normalized)
    np.testing.assert_allclose(out.unnormalized, expected_unnormalized)

  def test_learn_scale_shift(self):
    num_outputs = 2
    initial_state, update = pop_art.popart(
        num_outputs, step_size=1e-1, scale_lb=1e-6, scale_ub=1e6)
    state = initial_state()
    params = get_constant_linear_params(num_outputs)
    targets = np.arange(6) - 3
    indices = np.asarray([0, 0, 0, 1, 1, 1])
    # Learn the parameters.
    for _ in range(10):
      _, state = update(params, state, targets, indices)

    expected_scale = np.std(targets[:3])
    expected_scale = np.asarray([expected_scale, expected_scale])
    expected_shift = np.asarray([-2., 1.])
    # Loose tolerances; just get close.
    np.testing.assert_allclose(
        state.scale, expected_scale, atol=1e-1, rtol=1e-1)
    np.testing.assert_allclose(
        state.shift, expected_shift, atol=1e-1, rtol=1e-1)

  def test_slow_update(self):
    num_outputs = 2
    # Two step sizes: 0.1, and 0.8
    kwargs = dict(
        num_outputs=num_outputs,
        scale_lb=1e-6,
        scale_ub=1e6,
    )
    initial_state, slow_update = pop_art.popart(step_size=1e-2, **kwargs)
    _, fast_update = pop_art.popart(step_size=1e-1, **kwargs)
    state = initial_state()
    params = get_constant_linear_params(num_outputs)
    targets = np.arange(6) * 3  # standard deviation > 1 and mean > 0
    indices = np.asarray([0, 0, 0, 1, 1, 1])
    _, slow_state = slow_update(params, state, targets, indices)
    _, fast_state = fast_update(params, state, targets, indices)

    # Faster step size means faster adjustment.
    np.testing.assert_array_less(slow_state.shift, fast_state.shift)
    np.testing.assert_array_less(slow_state.scale, fast_state.scale)

  def test_scale_bounded(self):
    num_outputs = 1
    # Set scale_lb and scale_ub to 1 and verify this is obeyed.
    initial_state, update = pop_art.popart(
        num_outputs, step_size=1e-1, scale_lb=1., scale_ub=1.)
    state = initial_state()
    params = get_constant_linear_params(num_outputs)
    targets = np.ones((4, 2))
    indices = np.zeros((4, 2), dtype=np.int32)
    for _ in range(4):
      _, state = update(params, state, targets, indices)
      self.assertAlmostEqual(float(state.scale[0]), 1.)

  def test_outputs_preserved(self):
    num_outputs = 2
    initial_state, update = pop_art.popart(
        num_outputs, step_size=1e-3, scale_lb=1e-6, scale_ub=1e6)
    state = initial_state()
    key = jax.random.PRNGKey(428)

    def net(x):
      linear = hk.Linear(
          num_outputs, b_init=initializers.RandomUniform(), name='head')
      return linear(x)

    init_fn, apply_fn = hk.without_apply_rng(hk.transform(net))
    key, subkey1, subkey2 = jax.random.split(key, 3)
    fixed_data = jax.random.uniform(subkey1, (4, 3))
    params = init_fn(subkey2, fixed_data)
    initial_result = apply_fn(params, fixed_data)
    indices = np.asarray([0, 1, 0, 1, 0, 1, 0, 1])
    # Repeatedly update state and verify that params still preserve outputs.
    for _ in range(30):
      key, subkey1, subkey2 = jax.random.split(key, 3)
      targets = jax.random.uniform(subkey1, (8,))
      linear_params, state = update(params['head'], state, targets, indices)
      params = data_structures.to_mutable_dict(params)
      params['head'] = linear_params

      # Apply updated linear transformation and unnormalize outputs.
      transform = apply_fn(params, fixed_data)
      out = jnp.broadcast_to(state.scale,
                             transform.shape) * transform + jnp.broadcast_to(
                                 state.shift, transform.shape)
      np.testing.assert_allclose(initial_result, out, atol=1e-2)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `tree_util.py`."""

from absl.testing import absltest
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import tree_util

NUM_NESTS = 5


class TreeUtilTest(absltest.TestCase):

  def test_tree_split_key(self):
    rng_key = jax.random.PRNGKey(42)
    tree_like = (1, (2, 3), {'a': 4})
    _, tree_keys = tree_util.tree_split_key(rng_key, tree_like)
    self.assertLen(jax.tree_leaves(tree_keys), 4)

  def test_tree_map_zipped(self):
    nests = [
        dict(a=jnp.zeros((1, 3)), b=jnp.zeros((1, 5)))] * NUM_NESTS
    nest_output = tree_util.tree_map_zipped(
        lambda *args: jnp.concatenate(args), nests)
    self.assertEqual(nest_output['a'].shape, (NUM_NESTS, 3))
    self.assertEqual(nest_output['b'].shape, (NUM_NESTS, 5))

  def test_tree_map_zipped_wrong_structure(self):
    nests = [
        dict(a=jnp.zeros((1, 3)), b=jnp.zeros((1, 5)))] * (NUM_NESTS - 1)
    nests.append(dict(c=jnp.zeros((1, 3))))  # add a non-matching nest
    with self.assertRaisesRegex(ValueError, 'must share the same tree'):
      tree_util.tree_map_zipped(
          lambda *args: jnp.concatenate(args), nests)

  def test_tree_map_zipped_empty(self):
    outputs = tree_util.tree_map_zipped(lambda *args: jnp.concatenate(args), [])
    self.assertEmpty(outputs)

  def test_select_true(self):
    on_true = ((jnp.zeros(3,),), jnp.zeros(4,))
    on_false = ((jnp.ones(3,),), jnp.ones(4,))
    output = tree_util.tree_select(True, on_true, on_false)
    chex.assert_tree_all_close(output, on_true)

  def test_select_false(self):
    on_true = ((jnp.zeros(3,),), jnp.zeros(4,))
    on_false = ((jnp.ones(3,),), jnp.ones(4,))
    output = tree_util.tree_select(False, on_true, on_false)
    chex.assert_tree_all_close(output, on_false)

  def test_tree_split_leaves(self):
    t = {
        'a0': np.zeros(3),
        'd': {
            'a1': np.arange(3),
            'a2': np.zeros([3, 3]) + 2,
        },
        't4': (np.zeros([3, 2]) + 4, np.arange(3)),
    }

    for keepdim in (False, True):
      expd_shapes = jax.tree_map(lambda x: np.zeros(x.shape[1:]), t)
      if keepdim:
        expd_shapes = jax.tree_map(lambda x: np.expand_dims(x, 0), expd_shapes)

      res_trees = tree_util.tree_split_leaves(t, axis=0, keepdim=keepdim)
      self.assertLen(res_trees, 3)
      chex.assert_tree_all_equal_shapes(expd_shapes, *res_trees)
      for i, res_t in enumerate(res_trees):
        np.testing.assert_allclose(res_t['a0'], 0)
        np.testing.assert_allclose(res_t['d']['a1'], i)
        np.testing.assert_allclose(res_t['d']['a2'], 2)
        np.testing.assert_allclose(res_t['t4'][0], 4)
        np.testing.assert_allclose(res_t['t4'][1], i)

  def test_tree_fn(self):
    def add(x, y):
      return x + y
    tree_add = tree_util.tree_fn(add)
    tree_x = [1, 2, 3]
    tree_y = [1, 10, 100]
    output = tree_add(tree_x, tree_y)
    exp_output = [2, 12, 103]
    # Test output.
    np.testing.assert_allclose(output, exp_output)

  def test_tree_fn_with_partials(self):
    def add(x, y):
      return x + y
    tree_add = tree_util.tree_fn(add, y=1)
    tree_x = [1, 2, 3]
    output = tree_add(tree_x)
    exp_output = [2, 3, 4]
    # Test output.
    np.testing.assert_allclose(output, exp_output)

  def test_tree_replace_masked(self):
    data = jnp.array([
        [1, 2, 3, 4, 5, 6],
        [-1, -2, -3, -4, -5, -6]
    ])
    tree_data = {'a': data, 'b': data}
    replacement = data * 10
    tree_replacement = {'a': replacement, 'b': replacement}
    mask = jnp.array([0, 1])
    tree_output = tree_util.tree_replace_masked(
        tree_data, tree_replacement, mask)
    expected_output = jnp.array([
        [1, 2, 3, 4, 5, 6],
        [-10, -20, -30, -40, -50, -60],
    ])
    expected_tree_output = {'a': expected_output, 'b': expected_output}
    # Test output.
    np.testing.assert_allclose(tree_output['a'], expected_tree_output['a'])
    np.testing.assert_allclose(tree_output['b'], expected_tree_output['b'])

  def test_transpose_last_axis_to_first(self):
    tree = {'a': jnp.ones((1, 2, 3, 4)), 'b': {'c': jnp.ones((1, 2, 3, 4))}}
    expected_tree = {
        'a': jnp.ones((4, 1, 2, 3)),
        'b': {
            'c': jnp.ones((4, 1, 2, 3))
        }
    }
    chex.assert_trees_all_equal_shapes(
        tree_util.transpose_last_axis_to_first(tree['a']), expected_tree['a'])
    chex.assert_trees_all_equal_shapes(
        tree_util.transpose_last_axis_to_first(tree), expected_tree)

  def test_transpose_first_axis_to_last(self):
    tree = {'a': jnp.ones((1, 2, 3, 4)), 'b': {'c': jnp.ones((1, 2, 3, 4))}}
    expected_tree = {
        'a': jnp.ones((2, 3, 4, 1)),
        'b': {
            'c': jnp.ones((2, 3, 4, 1))
        }
    }
    chex.assert_trees_all_equal_shapes(
        tree_util.transpose_first_axis_to_last(tree['a']), expected_tree['a'])
    chex.assert_trees_all_equal_shapes(
        tree_util.transpose_first_axis_to_last(tree), expected_tree)

if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An online Q-lambda agent trained to play BSuite's Catch env."""

import collections
from absl import app
from absl import flags
from bsuite.environments import catch
import dm_env
import haiku as hk
from haiku import nets
import jax
import jax.numpy as jnp
import numpy as np
import optax
import rlax
from rlax.examples import experiment

ActorOutput = collections.namedtuple("ActorOutput", ["actions", "q_values"])

FLAGS = flags.FLAGS
flags.DEFINE_integer("seed", 42, "Random seed.")
flags.DEFINE_integer("train_episodes", 500, "Number of train episodes.")
flags.DEFINE_integer("num_hidden_units", 50, "Number of network hidden units.")
flags.DEFINE_integer("sequence_length", 4,
                     "Length of (action, timestep) sequences.")
flags.DEFINE_float("epsilon", 0.01, "Epsilon-greedy exploration probability.")
flags.DEFINE_float("lambda_", 0.9, "Mixing parameter for Q(lambda).")
flags.DEFINE_float("discount_factor", 0.99, "Q-learning discount factor.")
flags.DEFINE_float("learning_rate", 0.005, "Optimizer learning rate.")
flags.DEFINE_integer("eval_episodes", 100, "Number of evaluation episodes.")
flags.DEFINE_integer("evaluate_every", 50,
                     "Number of episodes between evaluations.")


def build_network(num_hidden_units: int, num_actions: int) -> hk.Transformed:
  """Factory for a simple MLP network for approximating Q-values."""

  def q(obs):
    flatten = lambda x: jnp.reshape(x, (-1,))
    network = hk.Sequential(
        [flatten, nets.MLP([num_hidden_units, num_actions])])
    return network(obs)

  return hk.without_apply_rng(hk.transform(q))


class SequenceAccumulator:
  """Accumulator for gathering the latest timesteps into sequences.

  Note sequences can overlap and cross episode boundaries.
  """

  def __init__(self, length):
    self._timesteps = collections.deque(maxlen=length)

  def push(self, timestep, action):
    # Replace `None`s with zeros as these will be put into NumPy arrays.
    a_tm1 = 0 if action is None else action
    timestep_t = timestep._replace(
        step_type=int(timestep.step_type),
        reward=0. if timestep.reward is None else timestep.reward,
        discount=0. if timestep.discount is None else timestep.discount,
    )
    self._timesteps.append((a_tm1, timestep_t))

  def sample(self, batch_size):
    """Returns current sequence of accumulated timesteps."""
    if batch_size != 1:
      raise ValueError("Require batch_size == 1.")
    if len(self._timesteps) != self._timesteps.maxlen:
      raise ValueError("Not enough timesteps for a full sequence.")

    actions, timesteps = jax.tree_map(lambda *ts: np.stack(ts),
                                           *self._timesteps)
    return actions, timesteps

  def is_ready(self, batch_size):
    if batch_size != 1:
      raise ValueError("Require batch_size == 1.")
    return len(self._timesteps) == self._timesteps.maxlen


class OnlineQLambda:
  """An online Q-lambda agent."""

  def __init__(self, observation_spec, action_spec, num_hidden_units, epsilon,
               lambda_, learning_rate):
    self._observation_spec = observation_spec
    self._action_spec = action_spec
    self._epsilon = epsilon
    self._lambda = lambda_

    # Neural net and optimiser.
    self._network = build_network(num_hidden_units, action_spec.num_values)

    self._optimizer = optax.adam(learning_rate)
    # Jitting for speed.
    self.actor_step = jax.jit(self.actor_step)
    self.learner_step = jax.jit(self.learner_step)

  def initial_params(self, key):
    sample_input = self._observation_spec.generate_value()
    return self._network.init(key, sample_input)

  def initial_actor_state(self):
    return ()

  def initial_learner_state(self, params):
    return self._optimizer.init(params)

  def actor_step(self, params, env_output, actor_state, key, evaluation):
    q = self._network.apply(params, env_output.observation)
    train_a = rlax.epsilon_greedy(self._epsilon).sample(key, q)
    eval_a = rlax.greedy().sample(key, q)
    a = jax.lax.select(evaluation, eval_a, train_a)
    return ActorOutput(actions=a, q_values=q), actor_state

  def learner_step(self, params, data, learner_state, unused_key):
    dloss_dtheta = jax.grad(self._loss)(params, *data)
    updates, learner_state = self._optimizer.update(dloss_dtheta, learner_state)
    params = optax.apply_updates(params, updates)
    return params, learner_state

  def _loss(self, params, actions, timesteps):
    """Calculates Q-lambda loss given parameters, actions and timesteps."""
    network_apply_sequence = jax.vmap(self._network.apply, in_axes=(None, 0))
    q = network_apply_sequence(params, timesteps.observation)

    # Use a mask since the sequence could cross episode boundaries.
    mask = jnp.not_equal(timesteps.step_type, int(dm_env.StepType.LAST))
    a_tm1 = actions[1:]
    r_t = timesteps.reward[1:]
    # Discount ought to be zero on a LAST timestep, use the mask to ensure this.
    discount_t = timesteps.discount[1:] * mask[1:]
    q_tm1 = q[:-1]
    q_t = q[1:]
    mask_tm1 = mask[:-1]

    # Mask out TD errors for the last state in an episode.
    td_error_tm1 = mask_tm1 * rlax.q_lambda(
        q_tm1, a_tm1, r_t, discount_t, q_t, lambda_=self._lambda)
    return jnp.sum(rlax.l2_loss(td_error_tm1)) / jnp.sum(mask_tm1)


def main(unused_arg):
  env = catch.Catch(seed=FLAGS.seed)
  agent = OnlineQLambda(
      observation_spec=env.observation_spec(),
      action_spec=env.action_spec(),
      num_hidden_units=FLAGS.num_hidden_units,
      epsilon=FLAGS.epsilon,
      lambda_=FLAGS.lambda_,
      learning_rate=FLAGS.learning_rate)

  accumulator = SequenceAccumulator(length=FLAGS.sequence_length)
  experiment.run_loop(
      agent=agent,
      environment=env,
      accumulator=accumulator,
      seed=FLAGS.seed,
      batch_size=1,
      train_episodes=FLAGS.train_episodes,
      evaluate_every=FLAGS.evaluate_every,
      eval_episodes=FLAGS.eval_episodes,
  )


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Unit tests for `transforms.py`."""

import functools
from absl.testing import absltest
from absl.testing import parameterized
import chex
import jax
import jax.numpy as jnp
import numpy as np
from rlax._src import transforms


TWO_HOT_BINS = 5
TWO_HOT_SCALARS = [-5.0, -3.0, -1.0, -0.4, 0.0, 0.3, 1.0, 4.5, 10.0]
TWO_HOT_PROBABILITIES = [
    [1.0, 0.0, 0.0, 0.0, 0.0],
    [1.0, 0.0, 0.0, 0.0, 0.0],
    [1.0, 0.0, 0.0, 0.0, 0.0],
    [0.0, 0.8, 0.2, 0.0, 0.0],
    [0.0, 0.0, 1.0, 0.0, 0.0],
    [0.0, 0.0, 0.4, 0.6, 0.0],
    [0.0, 0.0, 0.0, 0.0, 1.0],
    [0.0, 0.0, 0.0, 0.0, 1.0],
    [0.0, 0.0, 0.0, 0.0, 1.0]
]


class TransformsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.x = 0.5
    self.xs = jnp.array([-2, -1, -0.5, 0, 0.5, 1, 2])

  @chex.all_variants()
  def test_identity_scalar(self):
    identity = self.variant(transforms.identity)
    x = jnp.array(self.x)
    # Test output.
    np.testing.assert_allclose(identity(x), self.x)

  @chex.all_variants()
  def test_identity_vector(self):
    identity = self.variant(transforms.identity)
    # Test output.
    np.testing.assert_allclose(identity(self.xs), self.xs)

  @chex.all_variants()
  def test_sigmoid_scalar(self):
    sigmoid = self.variant(transforms.sigmoid)
    logit = self.variant(transforms.logit)
    x = jnp.array(self.x)
    # Test output.
    np.testing.assert_allclose(logit(sigmoid(x)), self.x, atol=1e-3)

  @chex.all_variants()
  def test_sigmoid_vector(self):
    sigmoid = self.variant(transforms.sigmoid)
    logit = self.variant(transforms.logit)
    # Test output.
    np.testing.assert_allclose(logit(sigmoid(self.xs)), self.xs, atol=1e-3)

  @chex.all_variants()
  def test_signed_log_exp_transform_scalar(self):
    signed_logp1 = self.variant(transforms.signed_logp1)
    signed_expm1 = self.variant(transforms.signed_expm1)
    x = jnp.array(self.x)
    # Test inverse.
    np.testing.assert_allclose(signed_expm1(signed_logp1(x)), self.x, atol=1e-3)

  @chex.all_variants()
  def test_signed_log_exp_transform_vector(self):
    signed_logp1 = self.variant(transforms.signed_logp1)
    signed_expm1 = self.variant(transforms.signed_expm1)
    # Test inverse.
    np.testing.assert_allclose(
        signed_expm1(signed_logp1(self.xs)), self.xs, atol=1e-3)

  @chex.all_variants()
  def test_signed_hyper_parabolic_transform_scalar(self):
    signed_hyperbolic = self.variant(transforms.signed_hyperbolic)
    signed_parabolic = self.variant(transforms.signed_parabolic)
    x = jnp.array(self.x)
    # Test inverse.
    np.testing.assert_allclose(
        signed_parabolic(signed_hyperbolic(x)), self.x, atol=1e-3)

  @chex.all_variants()
  def test_signed_hyper_parabolic_transform_vector(self):
    signed_hyperbolic = self.variant(transforms.signed_hyperbolic)
    signed_parabolic = self.variant(transforms.signed_parabolic)
    # Test inverse.
    np.testing.assert_allclose(
        signed_parabolic(signed_hyperbolic(self.xs)), self.xs, atol=1e-3)

  @chex.all_variants()
  def test_signed_power_transform_scalar(self):
    square = self.variant(functools.partial(transforms.power, p=2.))
    sqrt = self.variant(functools.partial(transforms.power, p=1/2.))
    x = jnp.array(self.x)
    # Test inverse.
    np.testing.assert_allclose(square(sqrt(x)), self.x, atol=1e-3)

  @chex.all_variants()
  def test_signed_power_transform_vector(self):
    square = self.variant(functools.partial(transforms.power, p=2.))
    sqrt = self.variant(functools.partial(transforms.power, p=1/2.))
    # Test inverse.
    np.testing.assert_allclose(square(sqrt(self.xs)), self.xs, atol=1e-3)

  @chex.all_variants()
  def test_hyperbolic_sin_transform_scalar(self):
    sinh = self.variant(transforms.hyperbolic_sin)
    arcsinh = self.variant(transforms.hyperbolic_arcsin)
    x = jnp.array(self.x)
    # Test inverse.
    np.testing.assert_allclose(sinh(arcsinh(x)), self.x, atol=1e-3)
    np.testing.assert_allclose(arcsinh(sinh(x)), self.x, atol=1e-3)

  @chex.all_variants()
  def test_hyperbolic_sin_transform_vector(self):
    sinh = self.variant(transforms.hyperbolic_sin)
    arcsinh = self.variant(transforms.hyperbolic_arcsin)
    # Test inverse.
    np.testing.assert_allclose(sinh(arcsinh(self.xs)), self.xs, atol=1e-3)
    np.testing.assert_allclose(arcsinh(sinh(self.xs)), self.xs, atol=1e-3)

  def test_transform_to_2hot(self):
    y = transforms.transform_to_2hot(
        scalar=jnp.array(TWO_HOT_SCALARS),
        min_value=-1.0,
        max_value=1.0,
        num_bins=TWO_HOT_BINS)

    np.testing.assert_allclose(y, np.array(TWO_HOT_PROBABILITIES), atol=1e-4)

  def test_transform_from_2hot(self):
    y = transforms.transform_from_2hot(
        probs=jnp.array(TWO_HOT_PROBABILITIES),
        min_value=-1.0,
        max_value=1.0,
        num_bins=TWO_HOT_BINS)

    np.testing.assert_allclose(
        y, np.clip(np.array(TWO_HOT_SCALARS), -1, 1), atol=1e-4)

  def test_2hot_roundtrip(self):
    min_value = -1.0
    max_value = 1.0
    num_bins = 11

    value = np.arange(min_value, max_value, 0.01)

    transformed = transforms.transform_to_2hot(
        value, min_value, max_value, num_bins)
    restored = transforms.transform_from_2hot(
        transformed, min_value, max_value, num_bins)

    np.testing.assert_almost_equal(value, restored, decimal=5)


if __name__ == '__main__':
  jax.config.update('jax_numpy_rank_promotion', 'raise')
  absltest.main()
